import pytest

from remoulade.brokers.rabbitmq import ChannelPool
from remoulade.errors import ChannelPoolTimeout


def test_channel_pool_acquire(mock_channel_factory):
    channel_pool = ChannelPool(channel_factory=mock_channel_factory, pool_size=10)

    with channel_pool.acquire() as channel_1:
        assert channel_1.id == 0
        assert len(channel_pool) == 9

        with channel_pool.acquire() as channel_2:
            assert channel_2.id == 1
            assert len(channel_pool) == 8

    assert len(channel_pool) == 10
    items_in_pool = []
    while len(channel_pool):
        channel = channel_pool.get()
        if channel is not None:
            items_in_pool.append(channel.id)
        else:
            items_in_pool.append(channel)

    assert items_in_pool == [0, 1, None, None, None, None, None, None, None, None]


def test_channel_pool_acquire_one_used(mock_channel_factory):
    channel_pool = ChannelPool(channel_factory=mock_channel_factory, pool_size=10)

    for _ in range(10):
        with channel_pool.acquire() as channel_1:
            assert channel_1.id == 0
            assert len(channel_pool) == 9

    assert len(channel_pool) == 10
    items_in_pool = []
    while len(channel_pool):
        channel = channel_pool.get()
        if channel is not None:
            items_in_pool.append(channel.id)
        else:
            items_in_pool.append(channel)

    assert items_in_pool == [0, None, None, None, None, None, None, None, None, None]


def test_raise_channel_pool_timeout(mock_channel_factory):
    channel_pool = ChannelPool(channel_factory=mock_channel_factory, pool_size=1)
    with channel_pool.acquire():
        assert len(channel_pool) == 0
        with pytest.raises(ChannelPoolTimeout):
            with channel_pool.acquire(timeout=1):
                pass


def test_channel_pool_clear(mock_channel_factory):
    channel_pool = ChannelPool(channel_factory=mock_channel_factory, pool_size=5)
    with channel_pool.acquire():
        with channel_pool.acquire():
            with channel_pool.acquire():
                with channel_pool.acquire():
                    with channel_pool.acquire():
                        pass

    channel_pool.clear()
    items_in_pool = []
    while len(channel_pool):
        items_in_pool.append(channel_pool.get())

    assert items_in_pool == [None, None, None, None, None]



import platform
import time
from typing import Dict, List
from unittest.mock import patch

import pytest

import remoulade
from remoulade import Message, Middleware
from remoulade.errors import RateLimitExceeded
from remoulade.middleware import SkipMessage
from remoulade.results import Results

from .common import get_logs, worker

_current_platform = platform.python_implementation()


def test_actors_can_be_defined(stub_broker):
    # Given that I've decorated a function with @actor
    @remoulade.actor
    def add(x, y):
        return x + y

    # I expect that function to become an instance of Actor
    assert isinstance(add, remoulade.Actor)


def test_actors_can_be_assigned_predefined_options(stub_broker):
    # Given that I have a stub broker with the retries middleware
    # If I define an actor with a max_retries number
    @remoulade.actor(max_retries=32)
    def add(x, y):
        return x + y

    # I expect the option to persist
    assert add.options["max_retries"] == 32


def test_actors_cannot_be_assigned_arbitrary_options(stub_broker):
    # Given that I have a stub broker
    # If I define an actor with a nonexistent option
    @remoulade.actor(invalid_option=32)
    def add(x, y):
        return x + y

    # I expect it to raise a ValueError when I add it to a broker
    with pytest.raises(ValueError):
        stub_broker.declare_actor(add)


def test_actors_can_be_named(stub_broker):
    # Given that I've decorated a function with @actor and named it explicitly
    @remoulade.actor(actor_name="foo")
    def add(x, y):
        return x + y

    # I expect the returned function to have that name
    assert add.actor_name == "foo"


def test_actors_can_be_assigned_custom_queues(stub_broker):
    # Given that I've decorated a function with @actor and given it an explicit queue
    @remoulade.actor(queue_name="foo")
    def foo():
        pass

    # I expect it to be able to be declared without error
    stub_broker.declare_actor(foo)

    # I expect the returned function to use that queue
    assert foo.queue_name == "foo"


def test_actors_fail_given_invalid_queue_names(stub_broker):
    # If I define an actor with an invalid queue name
    # I expect a ValueError to be raised
    with pytest.raises(ValueError):

        @remoulade.actor(queue_name="$2@!@#")
        def foo():
            pass


def test_actors_can_be_called(stub_broker):
    # Given that I have an actor
    @remoulade.actor
    def add(x, y):
        return x + y

    # If I call it directly,
    # I expect it to run synchronously
    assert add(1, 2) == 3


def test_actors_can_be_sent_messages(stub_broker):
    # Given that I have an actor
    @remoulade.actor
    def add(x, y):
        return x + y

    # And this actor is declared
    stub_broker.declare_actor(add)

    # If I send it a message,
    # I expect it to enqueue a message
    enqueued_message = add.send(1, 2)
    enqueued_message_data = stub_broker.queues["default"].get(timeout=1)
    assert enqueued_message == Message.decode(enqueued_message_data)


def test_actors_no_broker():
    # Given an actor
    @remoulade.actor
    def do_work():
        return 42

    # If I send it a message
    # I expect an error explaining it has no broker
    with pytest.raises(ValueError) as e:
        do_work.send()
    assert str(e.value) == "No broker is set, did you forget to call set_broker ?"


def test_actors_can_perform_work(stub_broker, stub_worker):
    # Given that I have a database
    database = {}

    # And an actor that can write data to that database
    @remoulade.actor
    def put(key, value):
        database[key] = value

    # And this actor is declared
    stub_broker.declare_actor(put)

    # If I send that actor many async messages
    for i in range(100):
        assert put.send("key-%s" % i, i)

    # Then join on the queue
    stub_broker.join(put.queue_name)
    stub_worker.join()

    # I expect the database to be populated
    assert len(database) == 100


def test_actors_can_perform_work_with_kwargs(stub_broker, stub_worker):
    # Given that I have a database
    results = []

    # And an actor
    @remoulade.actor
    def add(x, y):
        results.append(x + y)

    # And this actor is declared
    stub_broker.declare_actor(add)

    # If I send it a message with kwargs
    add.send(x=1, y=2)

    # Then join on the queue
    stub_broker.join(add.queue_name)
    stub_worker.join()

    # I expect the database to be populated
    assert results == [3]


def test_actors_do_not_retry_by_default(stub_broker, stub_worker, caplog):
    # Given that I have a database
    attempts = []

    # And an actor that fails every time
    @remoulade.actor()
    def do_work():
        attempts.append(1)
        raise RuntimeError("failure")

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send it a message
    do_work.send()

    # And join on the queue
    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # Then I expect 1 attempts to have occurred
    assert sum(attempts) == 1

    # I expect the fail log to be a error
    records = get_logs(caplog, "Failed to process message")
    assert len(records) == 1
    assert records[0].levelname == "ERROR"


def test_actors_retry_on_failure(stub_broker, stub_worker, caplog):
    # Given that I have a database
    failures, successes = [], []

    # And an actor that fails the first time it's called
    @remoulade.actor(max_retries=3, min_backoff=100, max_backoff=500)
    def do_work():
        if sum(failures) == 0:
            failures.append(1)
            raise RuntimeError("First failure.")
        else:
            successes.append(1)

    # test log

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # If I send it a message
    do_work.send()

    # Then join on the queue
    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # I expect successes
    assert sum(successes) == 1

    # I expect the fail log to be a warning if the message is retried
    records = get_logs(caplog, "Failed to process message")
    assert len(records) == 1
    assert records[0].levelname == "WARNING"


def test_actors_retry_a_max_number_of_times_on_failure(stub_broker, stub_worker):
    # Given that I have a database
    attempts = []

    # And an actor that fails every time
    @remoulade.actor(max_retries=3, min_backoff=1, backoff_strategy="constant")
    def do_work():
        attempts.append(1)
        raise RuntimeError("failure")

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send it a message
    do_work.send()

    # And join on the queue
    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # Then I expect 4 attempts to have occurred
    assert sum(attempts) == 4


def test_actors_retry_for_a_max_time(stub_broker, stub_worker):
    # Given that I have a database
    attempts = []

    # And an actor that fails every time
    @remoulade.actor(max_age=100, min_backoff=50, max_backoff=500)
    def do_work():
        attempts.append(1)
        raise RuntimeError("failure")

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send it a message
    do_work.send()

    # And join on the queue
    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # Then I expect at least one attempt to have occurred
    assert sum(attempts) >= 1


def test_actors_age_limit_exception(stub_broker, stub_worker, result_backend):
    @remoulade.actor(store_results=True)
    def do_sleep():
        time.sleep(0.3)

    @remoulade.actor(store_results=True, max_age=100, pipe_on_error=True)
    def do_raise():
        raise RuntimeError("failure")

    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And this actor is declared
    stub_broker.declare_actor(do_sleep)
    stub_broker.declare_actor(do_raise)

    p = remoulade.pipeline([do_sleep.message(), do_raise.message()])
    p.run()

    # And join on the queue
    stub_broker.join(do_sleep.queue_name)
    stub_broker.join(do_raise.queue_name)
    stub_worker.join()

    err = p.result.get(block=True, raise_on_error=False)

    assert "AgeLimitException" in err.message


@pytest.mark.skipif(_current_platform == "PyPy", reason="Time limits are not supported under PyPy.")
def test_actors_can_be_assigned_time_limits(stub_broker, stub_worker):
    # Given that I have a database
    attempts, successes = [], []

    # And an actor with a time limit
    @remoulade.actor(max_retries=0, time_limit=500)
    def do_work():
        attempts.append(1)
        time.sleep(1)
        time.sleep(1)
        successes.append(1)

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send it a message
    do_work.send()

    # And join on the queue
    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # Then I expect it to fail
    assert sum(attempts) == 1
    assert sum(successes) == 0


@pytest.mark.skipif(_current_platform == "PyPy", reason="Time limits are not supported under PyPy.")
def test_actor_messages_can_be_assigned_time_limits(stub_broker, stub_worker):
    # Given that I have a database
    attempts, successes = [], []

    # And an actor without an explicit time limit
    @remoulade.actor(max_retries=0)
    def do_work():
        attempts.append(1)
        time.sleep(2)
        successes.append(1)

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # If I send it a message with a custom time limit
    do_work.send_with_options(time_limit=1000)

    # Then join on the queue
    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # I expect it to fail
    assert sum(attempts) == 1
    assert sum(successes) == 0


def test_actors_can_be_assigned_message_age_limits(stub_broker):
    # Given that I have a database
    runs = []

    # And an actor whose messages have an age limit
    @remoulade.actor(max_age=100)
    def do_work():
        runs.append(1)

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send it a message
    do_work.send()

    # And wait for its age limit to pass
    time.sleep(0.1)

    # Then join on its queue
    with worker(stub_broker, worker_timeout=100) as stub_worker:
        stub_broker.join(do_work.queue_name)
        stub_worker.join()

        # I expect the message to have been skipped
        assert sum(runs) == 0


def test_actors_can_delay_messages_independent_of_each_other(stub_broker, stub_worker):
    # Given that I have a database
    results = []

    # And an actor that appends a number to the database
    @remoulade.actor
    def append(x):
        results.append(x)

    # And this actor is declared
    stub_broker.declare_actor(append)

    # If I send it a delayed message
    append.send_with_options(args=(1,), delay=1500)

    # And then another delayed message with a smaller delay
    append.send_with_options(args=(2,), delay=1000)

    # Then join on the queue
    stub_broker.join(append.queue_name)
    stub_worker.join()

    # I expect the latter message to have been run first
    assert results == [2, 1]


def test_messages_belonging_to_missing_actors_are_rejected(stub_broker, stub_worker):
    # Given that I have a broker without actors
    # If I send it a message
    message = Message(
        queue_name="some-queue",
        actor_name="some-actor",
        args=(),
        kwargs={},
        options={},
    )
    stub_broker.declare_queue("some-queue")
    stub_broker.enqueue(message)

    # Then join on the queue
    stub_broker.join("some-queue")
    stub_worker.join()

    # I expect the message to end up on the dead letter queue
    assert stub_broker.dead_letters == [message]


def test_before_and_after_signal_failures_are_ignored(stub_broker, stub_worker):
    # Given that I have a middleware that raises exceptions when it
    # tries to process messages.
    class BrokenMiddleware(Middleware):
        def before_process_message(self, broker, message):
            raise RuntimeError("before process message error")

        def after_process_message(self, broker, message, *, result=None, exception=None):
            raise RuntimeError("after process message error")

    # And a database
    database = []

    # And an actor that appends values to the database
    @remoulade.actor
    def append(x):
        database.append(x)

    # And this actor is declared
    stub_broker.declare_actor(append)

    # If add that middleware to my broker
    stub_broker.add_middleware(BrokenMiddleware())

    # And send my actor a message
    append.send(1)

    # Then join on the queue
    stub_broker.join(append.queue_name)
    stub_worker.join()

    # I expect the task to complete successfully
    assert database == [1]


def test_middleware_can_decide_to_skip_messages(stub_broker, stub_worker):
    # Given a middleware that skips all messages
    skipped_messages = []

    class SkipMiddleware(Middleware):
        def before_process_message(self, broker, message):
            raise SkipMessage()

        def after_skip_message(self, broker, message):
            skipped_messages.append(1)

    stub_broker.add_middleware(SkipMiddleware())

    # And an actor that keeps track of its calls
    calls = []

    @remoulade.actor
    def track_call():
        calls.append(1)

    # And this actor is declared
    stub_broker.declare_actor(track_call)

    # When I send that actor a message
    track_call.send()

    # And join on the broker and the worker
    stub_broker.join(track_call.queue_name)
    stub_worker.join()

    # Then I expect the call list to be empty
    assert sum(calls) == 0

    # And the skipped_messages list to contain one item
    assert sum(skipped_messages) == 1


def test_workers_can_be_paused(stub_broker, stub_worker):
    # Given a paused worker
    stub_worker.pause()

    # And an actor that keeps track of its calls
    calls = []

    @remoulade.actor
    def track_call():
        calls.append(1)

    # And this actor is declared
    stub_broker.declare_actor(track_call)

    # When I send that actor a message
    track_call.send()

    # And wait for 100ms
    time.sleep(0.1)

    # Then no calls should be made
    assert calls == []

    # When I resume the worker and join on it
    stub_worker.resume()
    stub_broker.join(track_call.queue_name)
    stub_worker.join()

    # Then one call should be made
    assert calls == [1]


def test_actors_can_prioritize_work(stub_broker):
    with worker(stub_broker, worker_timeout=100, worker_threads=1) as stub_worker:
        # Given that I a paused worker
        stub_worker.pause()

        # And actors with different priorities
        calls = []

        @remoulade.actor(priority=10)
        def hi():
            calls.append("hi")

        @remoulade.actor(priority=0)
        def lo():
            calls.append("lo")

        # And these actors are declared
        for actor in [hi, lo]:
            stub_broker.declare_actor(actor)

        # When I send both actors a nubmer of messages
        for _ in range(10):
            lo.send()
            hi.send()

        # Then resume the worker and join on the queue
        stub_worker.resume()
        stub_broker.join(lo.queue_name)
        stub_worker.join()

        # Then the high priority actor should run first
        assert calls == ["hi"] * 10 + ["lo"] * 10


def test_actors_can_conditionally_retry(stub_broker, stub_worker):
    # Given that I have a retry predicate
    def should_retry(retry_count, exception):
        return retry_count < 3 and isinstance(exception, RuntimeError)

    # And an actor that raises different types of errors
    attempts = []

    @remoulade.actor(retry_when=should_retry, min_backoff=100, max_backoff=100)
    def raises_errors(raise_runtime_error):
        attempts.append(1)
        if raise_runtime_error:
            raise RuntimeError("Runtime error")
        raise ValueError("Value error")

    # And this actor is declared
    stub_broker.declare_actor(raises_errors)

    # When I send that actor a message that makes it raise a value error
    raises_errors.send(False)

    # And wait for it
    stub_broker.join(raises_errors.queue_name)
    stub_worker.join()

    # Then I expect the actor not to retry
    assert sum(attempts) == 1

    # When I send that actor a message that makes it raise a runtime error
    attempts[:] = []
    raises_errors.send(True)

    # And wait for it
    stub_broker.join(raises_errors.queue_name)
    stub_worker.join()

    # Then I expect the actor to retry 3 times
    assert sum(attempts) == 4


def test_should_retry_with_max_retries(stub_broker, stub_worker):
    # Given that I have a retry predicate
    def should_retry(retry_count, exception):
        return isinstance(exception, RuntimeError)

    # And an actor that raises different types of errors
    attempts = []

    @remoulade.actor(retry_when=should_retry, max_retries=2, min_backoff=100, max_backoff=100)
    def raises_errors(raise_runtime_error):
        attempts.append(1)
        if raise_runtime_error:
            raise RuntimeError("Runtime error")
        raise ValueError("Value error")

    # And this actor is declared
    stub_broker.declare_actor(raises_errors)

    # When I send that actor a message that makes it raise a value error
    raises_errors.send(False)

    # And wait for it
    stub_broker.join(raises_errors.queue_name)
    stub_worker.join()

    # Then I expect the actor not to retry
    assert sum(attempts) == 1

    # When I send that actor a message that makes it raise a runtime error
    attempts[:] = []
    raises_errors.send(True)

    # And wait for it
    stub_broker.join(raises_errors.queue_name)
    stub_worker.join()

    # Then I expect the actor to retry twice
    assert sum(attempts) == 3


def test_can_call_str_on_actors():
    # Given that I have an actor
    @remoulade.actor
    def test():
        pass

    # When I call str on it
    # Then I should get back its representation as a string
    assert str(test) == "Actor(test)"


def test_can_call_repr_on_actors():
    # Given that I have an actor
    @remoulade.actor
    def test():
        pass

    # When I call repr on it
    # Then I should get back its representation
    assert repr(test) == "Actor(%(fn)r, queue_name='default', actor_name='test')" % vars(test)


def test_workers_log_rate_limit_exceeded_errors_differently(stub_broker, stub_worker):
    # Given that I've mocked the logging class
    with patch("logging.Logger.warning") as warning_mock:
        # And I have an actor that raises RateLimitExceeded
        @remoulade.actor(max_retries=0)
        def raise_rate_limit_exceeded():
            raise RateLimitExceeded("exceeded")

        # And this actor is declared
        stub_broker.declare_actor(raise_rate_limit_exceeded)

        # When I send that actor a message
        raise_rate_limit_exceeded.send()

        # And wait for the message to get processed
        stub_broker.join(raise_rate_limit_exceeded.queue_name)
        stub_worker.join()

        # Then warning mock should be called with a special message
        warning_messages = [args[0] for _, args, _ in warning_mock.mock_calls]
        assert "Rate limit exceeded in message %s: %s." in warning_messages


def test_as_dict_actor(stub_broker):
    @remoulade.actor
    def do_work(arg):
        return 1

    res = do_work.as_dict()
    assert res == {
        "args": [{"name": "arg"}],
        "name": "do_work",
        "priority": 0,
        "alternative_queues": None,
        "queue_name": "default",
    }


def test_as_dict_default(stub_broker):
    @remoulade.actor
    def do_work(arg=0):
        return 1

    res = do_work.as_dict()
    assert res == {
        "args": [{"default": "0", "name": "arg"}],
        "name": "do_work",
        "priority": 0,
        "queue_name": "default",
        "alternative_queues": None,
    }


def test_as_dict_typing(stub_broker):
    @remoulade.actor
    def do_work(a: List[int], b: Dict[str, bool], c: Dict[str, List[float]]):
        return 1

    res = do_work.as_dict()
    assert res == {
        "args": [
            {"name": "a", "type": "typing.List[int]"},
            {"name": "b", "type": "typing.Dict[str, bool]"},
            {"name": "c", "type": "typing.Dict[str, typing.List[float]]"},
        ],
        "name": "do_work",
        "priority": 0,
        "queue_name": "default",
        "alternative_queues": None,
    }


def test_as_dict_args(stub_broker):
    @remoulade.actor
    def do_work(*args):
        return 1

    res = do_work.as_dict()
    assert res == {
        "args": [{"name": "args"}],
        "name": "do_work",
        "priority": 0,
        "queue_name": "default",
        "alternative_queues": None,
    }


def test_as_dict_kwargs(stub_broker):
    @remoulade.actor
    def do_work(**kwargs):
        return 1

    res = do_work.as_dict()
    assert res == {
        "args": [{"name": "kwargs"}],
        "name": "do_work",
        "priority": 0,
        "queue_name": "default",
        "alternative_queues": None,
    }


def test_choose_queue(stub_broker):
    @remoulade.actor(queue_name="first_queue", alternative_queues=["second_queue"])
    def do_work():
        return 1

    stub_broker.declare_actor(do_work)

    msg = do_work.message_with_options(queue_name="second_queue")

    assert msg.queue_name == "second_queue"


def test_error_when_invalid_queue(stub_broker):
    @remoulade.actor
    def do_work():
        return 1

    stub_broker.declare_actor(do_work)

    with pytest.raises(ValueError):
        do_work.message_with_options(queue_name="invalid_queue")



import remoulade
from remoulade.helpers import compute_backoff, get_actor_arguments
from remoulade.helpers.reduce import reduce
from remoulade.results import Results


def test_reduce_messages(stub_broker, stub_worker, result_backend):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        return 1

    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def merge(results):
        return sum(results)

    # And this actor is declared
    stub_broker.declare_actor(do_work)
    stub_broker.declare_actor(merge)

    merged_message = reduce((do_work.message() for _ in range(10)), merge)
    merged_message.run()

    result = merged_message.result.get(block=True)

    assert 10 == result


def test_actor_arguments():
    @remoulade.actor
    def do_work(a: int = None):
        return 1

    assert get_actor_arguments(do_work) == [{"default": "", "name": "a", "type": "int"}]


def test_compute_backoff_exponential():
    assert compute_backoff(4, min_backoff=10, jitter=False, max_backoff=50, max_retries=4) == (5, 50)
    assert compute_backoff(2, min_backoff=10, jitter=False, max_backoff=100, max_retries=4) == (3, 40)
    assert compute_backoff(5, min_backoff=10, jitter=False, max_backoff=100, max_retries=3) == (6, 40)


def test_compute_backoff_constant():
    for retry in range(5):
        assert compute_backoff(retry, min_backoff=10, jitter=False, backoff_strategy="constant") == (retry + 1, 10)


def test_compute_backoff_linear():
    assert compute_backoff(3, min_backoff=10, jitter=False, backoff_strategy="linear") == (4, 40)
    assert compute_backoff(4, min_backoff=10, jitter=False, backoff_strategy="linear") == (5, 50)


def test_compute_backoff_spread_linear():
    assert compute_backoff(
        2, min_backoff=10, max_backoff=210, max_retries=5, jitter=False, backoff_strategy="spread_linear"
    ) == (3, 110)
    assert compute_backoff(
        3, min_backoff=10, max_backoff=210, max_retries=5, jitter=False, backoff_strategy="spread_linear"
    ) == (4, 160)
    assert compute_backoff(
        3, min_backoff=10, max_backoff=410, max_retries=5, jitter=False, backoff_strategy="spread_linear"
    ) == (4, 310)


def test_compute_backoff_spread_exponential():
    assert compute_backoff(
        0, min_backoff=10, jitter=False, max_backoff=100, max_retries=2, backoff_strategy="spread_exponential"
    ) == (1, 10)
    assert compute_backoff(
        1, min_backoff=10, jitter=False, max_backoff=160, max_retries=3, backoff_strategy="spread_exponential"
    ) == (2, 40)
    assert compute_backoff(
        4, min_backoff=10, jitter=False, max_backoff=160, max_retries=3, backoff_strategy="spread_exponential"
    ) == (5, 160)



import threading
import time
from unittest import mock
from unittest.mock import patch

import pytest
import redis

import remoulade
from remoulade import CollectionResults, Result, Worker
from remoulade.brokers.rabbitmq import RabbitmqBroker
from remoulade.middleware import Retries
from remoulade.results import ErrorStored, ResultBackend, ResultMissing, Results, ResultTimeout
from remoulade.results.backend import ForgottenResult
from remoulade.results.backends import StubBackend
from tests.conftest import fast_backoff


@pytest.mark.parametrize("forget", [True, False])
@pytest.mark.parametrize("block", [True, False])
def test_actors_can_store_results(stub_broker, stub_worker, result_backend, forget, block):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    if not block:
        stub_broker.join(do_work.queue_name)
        stub_worker.join()

    result = message.result.get(block=block, forget=forget)
    assert isinstance(message.result, Result)

    # Then the result should be what the actor returned
    assert result == 42


def test_actors_store_result_with_actor_name(stub_broker, stub_worker, result_backend):  # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)
    message = do_work.send()
    # Wait message to be process
    message.result.get(block=True)

    raw_backend_result = result_backend._get(f"remoulade-results:{message.message_id}")
    assert raw_backend_result["actor_name"] == "do_work"


def test_store_results_can_be_set_at_message_level(stub_broker, stub_worker, result_middleware):
    # Given an actor that do not stores results
    @remoulade.actor(store_results=False)
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message with store_results
    message = do_work.send_with_options(store_results=True)

    # And wait for a result
    assert message.result.get(block=True) == 42


@pytest.mark.parametrize("forget", [True, False])
def test_retrieving_a_result_can_raise_result_missing(stub_broker, stub_worker, result_backend, forget):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that sleeps for a long time before it stores a result
    @remoulade.actor(store_results=True)
    def do_work():
        time.sleep(0.2)
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And get the result without blocking
    # Then a ResultMissing error should be raised
    with pytest.raises(ResultMissing):
        result_backend.get_result(message, forget=forget)


@pytest.mark.parametrize("forget", [True, False])
def test_retrieving_a_result_can_time_out(stub_broker, stub_worker, result_backend, forget):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that sleeps for a long time before it stores a result
    @remoulade.actor(store_results=True)
    def do_work():
        time.sleep(0.5)
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    # Then a ResultTimeout error should be raised
    with pytest.raises(ResultTimeout):
        result_backend.get_result(message.message_id, block=True, timeout=200, forget=forget)


@pytest.mark.parametrize("forget", [True, False])
def test_default_result_timeout_can_be_set(stub_broker, stub_worker, result_backend, forget):
    # Given a result backend
    # And a broker with the results middleware with a result backend with a short_default_timeout set
    stub_broker.add_middleware(Results(backend=result_backend))
    result_backend.default_timeout = 200

    # And an actor that sleeps for a long time before it stores a result
    @remoulade.actor(store_results=True)
    def do_work():
        time.sleep(0.5)
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    # Then a ResultTimeout error should be raised
    with pytest.raises(ResultTimeout):
        result_backend.get_result(message.message_id, block=True, forget=forget)


@pytest.mark.parametrize("forget", [True, False])
def test_messages_can_get_results_from_backend(stub_broker, stub_worker, result_backend, forget):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that stores a result
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    # Then I should get that result back
    assert message.result.get(block=True, forget=forget) == 42


def test_messages_results_can_get_results_from_backend(stub_broker, stub_worker, result_backend):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that stores a result
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And create a message result
    result = Result(message_id=message.message_id)

    # And wait for a result
    # Then I should get that result back
    assert result.get(block=True) == 42


def test_messages_can_get_results_from_inferred_backend(stub_broker, stub_worker, redis_result_backend):
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=redis_result_backend))

    # And an actor that stores a result
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    # Then I should get that result back
    assert message.result.get(block=True) == 42


def test_result_default_before_retries(stub_broker, result_backend, stub_worker):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    retries_index, results_index = None, None

    for i, middleware in enumerate(stub_broker.middleware):
        if isinstance(middleware, Retries):
            retries_index = i
        if isinstance(middleware, Results):
            results_index = i

    assert results_index is not None
    assert retries_index is not None
    # The Results middleware should be before the Retries middleware
    assert retries_index > results_index


@pytest.mark.parametrize("block", [True, False])
def test_raise_on_error(stub_broker, result_backend, stub_worker, block):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that store a result and fail
    @remoulade.actor(store_results=True)
    def do_work():
        raise ValueError()

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    if not block:
        stub_broker.join(do_work.queue_name)
        stub_worker.join()

    # It should raise an error
    with pytest.raises(ErrorStored) as e:
        message.result.get(block=block)
    assert str(e.value) == "ValueError()"


@pytest.mark.parametrize("block", [True, False])
def test_store_errors(stub_broker, result_backend, stub_worker, block):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that store a result and fail
    @remoulade.actor(store_results=True)
    def do_work():
        raise ValueError()

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    if not block:
        stub_broker.join(do_work.queue_name)
        stub_worker.join()

    # Then I should get an ErrorStored
    error_stored = message.result.get(block=block, raise_on_error=False)
    assert isinstance(error_stored, ErrorStored)
    assert str(error_stored) == "ValueError()"


def test_store_errors_after_no_more_retry(stub_broker, result_backend, stub_worker):
    # Given that I have a database
    failures = []

    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # Given an actor that stores results
    @remoulade.actor(max_retries=3, store_results=True, min_backoff=10, max_backoff=100)
    def do_work():
        failures.append(1)
        raise ValueError()

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message,
    message = do_work.send()

    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # I get an error
    with pytest.raises(Exception) as e:
        message.result.get(block=True)
    assert str(e.value) == "ValueError()"

    # all the retries have been made
    assert sum(failures) == 4


@pytest.mark.parametrize("block", [True, False])
def test_messages_can_get_results_and_forget(stub_broker, stub_worker, result_backend, block):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that stores a result
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    if not block:
        stub_broker.join(do_work.queue_name)
        stub_worker.join()

    # Then I should get that result back
    assert message.result.get(block=block, forget=True) == 42

    # If I ask again for the same result it should have been forgotten
    assert message.result.get() is None


@pytest.mark.parametrize("block", [True, False])
@pytest.mark.parametrize("forget", [True, False])
def test_redis_backend_keep_ttl_all_time(stub_broker, stub_worker, redis_result_backend, block, forget):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=redis_result_backend))

    # And an actor that stores a result with a result_ttl
    result_ttl = 100 * 1000

    @remoulade.actor(store_results=True, result_ttl=result_ttl)
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    if not block:
        stub_broker.join(do_work.queue_name)
        stub_worker.join()
        # The result should have a TTL in redis
        assert redis_result_backend.client.ttl(redis_result_backend.build_message_key(message.message_id)) > 0

    # Then I should get that result back
    assert message.result.get(block=block, forget=forget) == 42
    # The forgotten result should still have a TTL in redis
    assert redis_result_backend.client.ttl(redis_result_backend.build_message_key(message.message_id)) > 0


@pytest.mark.parametrize("error", [True, False])
def test_messages_can_get_completed(stub_broker, stub_worker, result_backend, error):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        if error:
            raise ValueError()
        else:
            return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    result = message.result
    # we can get the completion
    assert result.completed

    result.get(forget=True, raise_on_error=False)

    # even after a forget
    assert result.completed


def test_result_get_forget_not_store_if_no_result(stub_broker, stub_worker, result_backend):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    event = threading.Event()

    # And an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        event.wait(2)
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    result = message.result
    # If I get the result and forget it
    with pytest.raises(ResultMissing):
        result.get(forget=True)

    event.set()
    # It should not store a forgotten result if there is no key
    assert result.get(block=True) == 42


def test_error_cannot_be_serialized(stub_broker, stub_worker, result_middleware):
    # given a exception which is not serializable
    class UnserializableError(Exception):
        def __repr__(self):
            raise ValueError()

    # And an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        raise UnserializableError()

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # And wait for a result
    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # If I get the result, i will get an error
    with pytest.raises(ErrorStored) as e:
        message.result.get(raise_on_error=True)
    assert e.value.message == "Exception could not be serialized"


@mock.patch("remoulade.results.backend.compute_backoff", fast_backoff)
def test_retry_if_saving_result_fail(stub_broker, stub_worker):
    with patch.object(ResultBackend, "store_results") as mock_store_results:
        mock_store_results.side_effect = Exception("Cannot save result")
        middleware = Results(backend=StubBackend())
        stub_broker.add_middleware(middleware)

        attempts = []

        # And an actor that stores results
        @remoulade.actor(store_results=True)
        def do_work():
            attempts.append(1)

        # And this actor is declared
        stub_broker.declare_actor(do_work)

        # When I send that actor a message
        do_work.send()

        # And wait for a result
        stub_broker.join(do_work.queue_name)
        stub_worker.join()

        # The actor has been tried 4 times
        assert len(attempts) == 4


@pytest.mark.parametrize("block", [True, False])
@pytest.mark.parametrize("forget", [True, False])
@mock.patch("remoulade.results.backends.redis.compute_backoff", fast_backoff)
def test_redis_get_result_retry_when_fail(redis_result_backend, block, forget):
    with patch.object(redis_result_backend, "client") as mock_client:
        mock_client.pipeline.side_effect = redis.ConnectionError()
        mock_client.brpoplpush.side_effect = redis.ConnectionError()
        mock_client.rpoplpush.side_effect = redis.TimeoutError()

        with pytest.raises(redis.RedisError):
            redis_result_backend.get_result("message-id", block=block, forget=forget)

        if block:
            assert mock_client.brpoplpush.call_count == 4
        elif forget:
            assert mock_client.pipeline.call_count == 4
        else:
            assert mock_client.rpoplpush.call_count == 4


@mock.patch("remoulade.results.backends.redis.compute_backoff", fast_backoff)
def test_redis_get_result_still_return_result_if_forget_fails(redis_result_backend):
    with patch.object(redis_result_backend, "client") as mock_client:
        mock_client.brpoplpush.return_value = redis_result_backend.encoder.encode(ForgottenResult.asdict())
        mock_client.pipeline.side_effect = redis.ConnectionError()
        assert redis_result_backend.get_result("message-id", block=True, forget=True) is None
        assert mock_client.brpoplpush.call_count == 1
        assert mock_client.pipeline.call_count == 1


@pytest.mark.parametrize("forget", [True, False])
async def test_redis_async_get_results_with_forget(
    rabbitmq_broker: RabbitmqBroker,
    rabbitmq_worker: Worker,
    redis_result_backend,
    forget: bool,
):
    rabbitmq_broker.add_middleware(Results(backend=redis_result_backend))

    # And an actor that stores a result
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()
    # Wait to get results
    rabbitmq_broker.join(do_work.queue_name)
    result = await message.result.async_get(forget=forget)
    assert result == 42

    if forget:
        forgotten_result = await message.result.async_get(forget=forget)
        assert forgotten_result is None
    else:
        result = await message.result.async_get(forget=forget)
        assert result == 42


@pytest.mark.parametrize("forget", [True])
async def test_redis_async_get_results_with_forget_timeout(
    rabbitmq_broker: RabbitmqBroker,
    rabbitmq_worker: Worker,
    redis_result_backend,
    forget: bool,
):
    rabbitmq_broker.add_middleware(Results(backend=redis_result_backend))

    @remoulade.actor(store_results=True)
    def do_work():
        time.sleep(0.2)
        return 42

    rabbitmq_broker.declare_actor(do_work)
    message = do_work.send()
    rabbitmq_broker.join(do_work.queue_name)
    with pytest.raises(ResultTimeout):
        await message.result.async_get(forget=forget, timeout=1)


def test_completed_count_no_messages(stub_broker, result_middleware):
    result = CollectionResults([])
    assert result.completed_count == 0



import datetime
import json
from enum import Enum
from typing import Optional, Tuple, Union

import pytest
from _decimal import Decimal
from pydantic import BaseModel, ValidationError

import remoulade
from remoulade import ActorNotFound, Message
from remoulade.encoder import JSONEncoder, MessageData, PydanticEncoder
from remoulade.results import Results
from remoulade.results.backend import BackendResult


def test_set_encoder_sets_the_global_encoder(pickle_encoder):
    # Given that I've set a Pickle encoder as the global encoder
    # When I get the global encoder
    encoder = remoulade.get_encoder()

    # Then it should be the same as the encoder that was set
    assert encoder == pickle_encoder


def test_pickle_encoder(pickle_encoder, stub_broker, stub_worker):
    # Given that I've set a Pickle encoder as the global encoder
    # And I have an actor that adds a value to a db
    db = []

    @remoulade.actor
    def add_value(x):
        db.append(x)

    # And this actor is declared
    stub_broker.declare_actor(add_value)

    # When I send that actor a message
    add_value.send(1)

    # And wait on the broker and worker
    stub_broker.join(add_value.queue_name)
    stub_worker.join()

    # Then I expect the message to have been processed
    assert db == [1]


class MyEnum(Enum):
    val = "val"
    other = "other"


class MyFirstSchema(BaseModel):
    name: str
    address: Optional[str]
    birth_date: datetime.datetime
    enum: MyEnum


class MySecondSchema(BaseModel):
    id: str
    value: Decimal


class FirstOutputSchema(BaseModel):
    output: Union[str, int]
    error: Union[str, None] = None


class SecondOutputSchema(BaseModel):
    val: str


def tuple_to_list(dict_value: dict, key: str) -> dict:
    return {**dict_value, key: [args for args in dict_value[key]]}


@remoulade.actor(store_results=True)
def my_actor_tuple(
    input_1: MyFirstSchema,
    input_2: Optional[MySecondSchema] = None,
    input_3: Union[MyFirstSchema, MySecondSchema, None] = None,
) -> Tuple[FirstOutputSchema, Optional[SecondOutputSchema]]:
    return FirstOutputSchema(output=input_1.name), SecondOutputSchema(val=input_2.id if input_2 is not None else "2")


@remoulade.actor(store_results=True)
def my_actor(
    input_1: MyFirstSchema,
    input_2: Optional[MySecondSchema] = None,
    input_3: Union[MyFirstSchema, MySecondSchema, None] = None,
) -> FirstOutputSchema:
    return FirstOutputSchema(output=input_1.name)


@remoulade.actor(store_results=True)
def my_actor_none(
    input_1: MyFirstSchema,
    input_2: Optional[MySecondSchema] = None,
    input_3: Union[MyFirstSchema, MySecondSchema, None] = None,
) -> None:
    return


@pytest.fixture
def input_1() -> MyFirstSchema:
    return MyFirstSchema(name="aa", address=None, birth_date=datetime.datetime(2022, 1, 1), enum=MyEnum.val)


@pytest.fixture
def input_2() -> MySecondSchema:
    return MySecondSchema(id="33", value=Decimal("2"))


@pytest.fixture
def message_data_decoded(input_1: MyFirstSchema, input_2: MySecondSchema) -> MessageData:
    return Message(
        queue_name="default",
        actor_name="my_actor_tuple",
        args=(input_1,),
        kwargs={"input_2": input_2},
        options={},
        message_id="dce31cca-70eb-4a55-9b16-5fb6cdd43415",
        message_timestamp=68139918115,
    ).asdict()


@pytest.fixture
def message_data_encoded() -> bytes:
    return (
        b'{"queue_name": "default", '
        b'"actor_name": "my_actor_tuple", '
        b'"args": [{"name": "aa", "address": null, "birth_date": "2022-01-01T00:00:00", "enum": "val"}], '
        b'"kwargs": {"input_2": {"id": "33", "value": "2"}}, '
        b'"options": {}, '
        b'"message_id": "dce31cca-70eb-4a55-9b16-5fb6cdd43415", '
        b'"message_timestamp": 68139918115}'
    )


@pytest.fixture
def encoder(stub_broker, stub_worker, result_backend) -> PydanticEncoder:
    stub_broker.add_middleware(Results(backend=result_backend))
    stub_broker.declare_actor(my_actor_tuple)
    stub_broker.declare_actor(my_actor_none)
    stub_broker.declare_actor(my_actor)
    return PydanticEncoder()


@pytest.fixture
def encoder_with_fallback(stub_broker, stub_worker, result_backend) -> PydanticEncoder:
    stub_broker.add_middleware(Results(backend=result_backend))
    stub_broker.declare_actor(my_actor_tuple)
    return PydanticEncoder(fallback_encoder=JSONEncoder())


def test_encoder_message(encoder: PydanticEncoder, message_data_decoded: MessageData, message_data_encoded: bytes):
    encoded_result = encoder.encode(message_data_decoded)
    assert encoded_result == message_data_encoded

    decoded_result = encoder.decode(message_data_encoded)

    # Args tuple are assumed to become list in remoulade
    assert decoded_result == tuple_to_list(message_data_decoded, "args")

    assert encoder.decode(encoder.encode(message_data_decoded)) == tuple_to_list(message_data_decoded, "args")
    assert encoder.encode(encoder.decode(message_data_encoded)) == message_data_encoded


def test_message_unknown_actor(encoder: PydanticEncoder, message_data_encoded: bytes):
    message_json_decoded = json.loads(message_data_encoded.decode("utf-8"))
    message_json_decoded["actor_name"] = "titi"
    with pytest.raises(ActorNotFound):
        encoder.decode(json.dumps(message_json_decoded).encode("utf-8"))


def test_message_fallback_no_actor_name(encoder_with_fallback: PydanticEncoder, message_data_encoded: bytes):
    message_json_decoded = json.loads(message_data_encoded.decode("utf-8"))
    message_json_decoded["actor_name"] = "titi"
    decoded_result = encoder_with_fallback.decode(json.dumps(message_json_decoded).encode("utf-8"))
    # Do not raise and keep dict instead of schema
    assert decoded_result == tuple_to_list(message_json_decoded, "args")


def test_message_schema_not_matching(encoder: PydanticEncoder, message_data_encoded: bytes):
    message_json_decoded = json.loads(message_data_encoded.decode("utf-8"))
    message_json_decoded["args"] = [{"toto": "a"}]
    message_json_decoded["kwargs"]["input_2"] = {"val": "aaa"}
    with pytest.raises(ValidationError):
        encoder.decode(json.dumps(message_json_decoded).encode("utf-8"))


@pytest.fixture
def backend_result_decoded(input_1: MyFirstSchema, input_2: MySecondSchema) -> MessageData:
    return BackendResult(result=my_actor(input_1, input_2), error=None, forgot=True, actor_name="my_actor").asdict()


@pytest.fixture
def backend_result_decoded_tuple(input_1: MyFirstSchema, input_2: MySecondSchema) -> MessageData:
    return BackendResult(
        result=my_actor_tuple(input_1, input_2), error=None, forgot=True, actor_name="my_actor_tuple"
    ).asdict()


@pytest.fixture
def backend_result_decoded_none(input_1: MyFirstSchema, input_2: MySecondSchema) -> MessageData:
    return BackendResult(result=None, error=None, forgot=True, actor_name="my_actor_none").asdict()


@pytest.fixture
def backend_result_decoded_raise(input_1: MyFirstSchema, input_2: MySecondSchema) -> MessageData:
    return BackendResult(result=None, error="Exception", forgot=True, actor_name="my_actor").asdict()


@pytest.fixture
def backend_result_encoded_raise() -> bytes:
    return b'{"result": null, "error": "Exception", "forgot": true, "actor_name": "my_actor"}'


@pytest.fixture
def backend_result_encoded() -> bytes:
    return b'{"result": {"output": "aa", "error": null}, ' b'"error": null, "forgot": true, "actor_name": "my_actor"}'


@pytest.fixture
def backend_result_encoded_tuple() -> bytes:
    return (
        b'{"result": [{"output": "aa", "error": null}, {"val": "33"}], '
        b'"error": null, "forgot": true, "actor_name": "my_actor_tuple"}'
    )


@pytest.fixture
def backend_result_encoded_none() -> bytes:
    return b'{"result": null, "error": null, "forgot": true, "actor_name": "my_actor_none"}'


def test_encoder_result(encoder: PydanticEncoder, backend_result_decoded: MessageData, backend_result_encoded: bytes):
    encoded_value = encoder.encode(backend_result_decoded)
    assert encoded_value == backend_result_encoded

    decoded_result = encoder.decode(backend_result_encoded)

    assert decoded_result == backend_result_decoded

    assert encoder.decode(encoder.encode(backend_result_decoded)) == backend_result_decoded
    assert encoder.encode(encoder.decode(backend_result_encoded)) == backend_result_encoded


def test_encoder_result_when_raise(
    encoder: PydanticEncoder, backend_result_decoded_raise: MessageData, backend_result_encoded_raise: bytes
):
    encoded_value = encoder.encode(backend_result_decoded_raise)
    assert encoded_value == backend_result_encoded_raise

    decoded_result = encoder.decode(backend_result_encoded_raise)

    assert decoded_result == backend_result_decoded_raise

    assert encoder.decode(encoder.encode(backend_result_decoded_raise)) == backend_result_decoded_raise
    assert encoder.encode(encoder.decode(backend_result_encoded_raise)) == backend_result_encoded_raise


def test_encoder_result_tuple(
    encoder: PydanticEncoder, backend_result_decoded_tuple: MessageData, backend_result_encoded_tuple: bytes
):
    encoded_value = encoder.encode(backend_result_decoded_tuple)
    assert encoded_value == backend_result_encoded_tuple

    decoded_result = encoder.decode(backend_result_encoded_tuple)

    assert decoded_result == backend_result_decoded_tuple

    assert encoder.decode(encoder.encode(backend_result_decoded_tuple)) == backend_result_decoded_tuple
    assert encoder.encode(encoder.decode(backend_result_encoded_tuple)) == backend_result_encoded_tuple


def test_encoder_result_with_none(
    encoder: PydanticEncoder, backend_result_decoded_none: MessageData, backend_result_encoded_none: bytes
):
    encoded_value = encoder.encode(backend_result_decoded_none)
    assert encoded_value == backend_result_encoded_none

    decoded_result = encoder.decode(backend_result_encoded_none)

    assert decoded_result == backend_result_decoded_none

    assert encoder.decode(encoder.encode(backend_result_decoded_none)) == backend_result_decoded_none
    assert encoder.encode(encoder.decode(backend_result_encoded_none)) == backend_result_encoded_none


def test_backend_result_unknown_actor(encoder: PydanticEncoder, backend_result_encoded_tuple: bytes):
    backend_result_json_decoded = json.loads(backend_result_encoded_tuple.decode("utf-8"))
    backend_result_json_decoded["actor_name"] = "titi"
    with pytest.raises(ActorNotFound):
        encoder.decode(json.dumps(backend_result_json_decoded).encode("utf-8"))


def test_backend_result_schema_not_matching(encoder: PydanticEncoder, backend_result_encoded_tuple: bytes):
    backend_result_json_decoded = json.loads(backend_result_encoded_tuple.decode("utf-8"))
    backend_result_json_decoded["result"] = [{"val": "titi"}]
    with pytest.raises(ValidationError):
        encoder.decode(json.dumps(backend_result_json_decoded).encode("utf-8"))


def test_fallback_no_schema(encoder_with_fallback: PydanticEncoder, backend_result_encoded_tuple: bytes):
    backend_result_json_decoded = json.loads(backend_result_encoded_tuple.decode("utf-8"))
    backend_result_json_decoded["result"] = [{"val": "titi"}]
    decoded_backend_result = encoder_with_fallback.decode(json.dumps(backend_result_json_decoded).encode("utf-8"))
    # Do not raise and keep dict instead of schema
    assert decoded_backend_result == backend_result_json_decoded



import pytest

from remoulade.helpers.queues import dq_name, q_name, xq_name


@pytest.mark.parametrize(
    "given,expected", [("default", "default"), ("default.DQ", "default"), ("default.XQ", "default")]
)
def test_q_name_returns_canonical_names(given, expected):
    assert q_name(given) == expected


@pytest.mark.parametrize(
    "given,expected", [("default", "default.DQ"), ("default.DQ", "default.DQ"), ("default.XQ", "default.DQ")]
)
def test_dq_name_returns_canonical_delay_names(given, expected):
    assert dq_name(given) == expected


@pytest.mark.parametrize(
    "given,expected", [("default", "default.XQ"), ("default.DQ", "default.XQ"), ("default.XQ", "default.XQ")]
)
def test_xq_name_returns_delay_names(given, expected):
    assert xq_name(given) == expected



import datetime
import json
import threading
import time

import pytest
import pytz

import remoulade
from remoulade.scheduler import ScheduledJob
from tests.conftest import check_redis, mock_func, new_scheduler


def test_simple_interval_scheduler(stub_broker, stub_worker, scheduler, scheduler_thread, mul, add):
    result = 0

    @remoulade.actor()
    def write_loaded_at():
        nonlocal result
        result += 1

    stub_broker.declare_actor(write_loaded_at)
    write_loaded_at.send, event_write = mock_func(write_loaded_at.send)
    mul.send, event_mul = mock_func(mul.send)
    start = time.time()

    # Run scheduler
    scheduler.schedule = [
        ScheduledJob(
            actor_name="write_loaded_at",
            interval=1,
        ),
        ScheduledJob(actor_name="mul", kwargs={"x": 1, "y": 2}, interval=3600),
    ]
    scheduler_thread.start()

    event_write.wait(10)
    event_mul.wait(10)

    stub_broker.join(mul.queue_name)
    stub_broker.join(write_loaded_at.queue_name)
    stub_worker.join()

    end = time.time()
    # should have written ~1 line per second
    assert end - start - 1 <= result <= end - start + 1

    # get the last_queued date for this slow task, this should not change when reloading schedule with new config
    tasks = scheduler.get_redis_schedule().values()
    (slow_task,) = (job for job in tasks if job.actor_name == "mul")
    last_queued = slow_task.last_queued
    assert {j.actor_name for j in tasks} == {"mul", "write_loaded_at"}

    scheduler.schedule = [
        ScheduledJob(actor_name="add", kwargs={"x": 1, "y": 2}, interval=1),
        ScheduledJob(actor_name="mul", kwargs={"x": 1, "y": 2}, interval=3600),
    ]
    scheduler.sync_config()

    tasks = scheduler.get_redis_schedule().values()
    # One item was deleted
    assert {j.actor_name for j in tasks} == {"add", "mul"}
    # The other one was not updated
    (slow_task,) = (job for job in tasks if job.actor_name == "mul")
    assert slow_task.last_queued == last_queued


def test_multiple_schedulers(stub_broker, stub_worker):
    result = 0

    @remoulade.actor
    def write_loaded_at():
        nonlocal result
        result += 1

    stub_broker.declare_actor(write_loaded_at)
    schedule = [
        ScheduledJob(
            actor_name="write_loaded_at",
            interval=3600,
        )
    ]

    scheduler_list = []
    event_list = []
    thread_list = []

    for _ in range(5):
        sch = new_scheduler(stub_broker)
        sch.get_redis_schedule, event = mock_func(sch.get_redis_schedule)
        event_list.append(event)
        if not scheduler_list:
            check_redis(sch.client)
        sch.schedule = schedule
        scheduler_list.append(sch)
        t = threading.Thread(target=sch.start)
        thread_list.append(t)
        t.start()

    for _ in range(2):
        for event in event_list:
            event.wait(2)
            event.clear()

    stub_broker.join(write_loaded_at.queue_name)
    stub_worker.join()

    # slow task should run exactly once, even if we launched 2 schedulers
    assert result == 1

    for scheduler in scheduler_list:
        scheduler.stop()

    for thread in thread_list:
        thread.join(10)


@pytest.mark.parametrize("tz", [None, "Europe/Paris"])
def test_scheduler_daily_time(stub_broker, stub_worker, scheduler, scheduler_thread, tz, frozen_datetime):
    result = 0

    @remoulade.actor
    def write_loaded_at():
        nonlocal result
        result += 1

    stub_broker.declare_actor(write_loaded_at)
    scheduler.get_redis_schedule, event_sch = mock_func(scheduler.get_redis_schedule)
    write_loaded_at.send, event_send = mock_func(write_loaded_at.send)

    scheduler.schedule = [
        ScheduledJob(
            actor_name="write_loaded_at",
            daily_time=(
                datetime.datetime.now(pytz.timezone(tz) if tz else None) + datetime.timedelta(seconds=1)
            ).time(),
            tz=tz,
        )
    ]
    scheduler_thread.start()
    # Wait for sync_config + a complete scheduler iteration
    for _ in range(3):
        event_sch.wait(10)
        event_sch.clear()
    stub_broker.join(write_loaded_at.queue_name)
    stub_worker.join()
    # should not have run yet
    assert result == 0
    frozen_datetime.tick(2)
    # Wait for the ScheduledJob to be sent
    event_send.wait(1)
    stub_broker.join(write_loaded_at.queue_name)
    stub_worker.join()
    assert result == 1

    # Wait for a complete scheduler iteration
    for _ in range(2):
        event_sch.wait(10)
        event_sch.clear()
    stub_broker.join(write_loaded_at.queue_name)
    stub_worker.join()

    # should not rerun
    assert result == 1


def test_scheduler_new_daily_time(stub_broker, stub_worker, scheduler, scheduler_thread):
    result = 0

    @remoulade.actor
    def write_loaded_at():
        nonlocal result
        result += 1

    stub_broker.declare_actor(write_loaded_at)

    scheduler.schedule = [
        ScheduledJob(
            actor_name="write_loaded_at",
            daily_time=(datetime.datetime.utcnow() - datetime.timedelta(seconds=1)).time(),
        )
    ]
    scheduler.get_redis_schedule, event = mock_func(scheduler.get_redis_schedule)
    scheduler_thread.start()
    event.wait(3)
    event.clear()
    event.wait(2)

    stub_broker.join(write_loaded_at.queue_name)
    stub_worker.join()

    # should not have ran, will run tomorrow
    assert result == 0


def test_scheduler_wrong_weekday(stub_broker, stub_worker, scheduler, scheduler_thread):
    result = 0

    @remoulade.actor
    def write_loaded_at():
        nonlocal result
        result += 1

    stub_broker.declare_actor(write_loaded_at)

    scheduler.schedule = [
        ScheduledJob(
            actor_name="write_loaded_at",
            iso_weekday=datetime.datetime.now().isoweekday() + 1,
        )
    ]
    scheduler.get_redis_schedule, event = mock_func(scheduler.get_redis_schedule)
    scheduler_thread.start()

    event.wait(2)
    event.clear()
    event.wait(2)

    # do nothing
    assert result == 0


def test_scheduler_right_weekday(stub_broker, stub_worker, scheduler, scheduler_thread):
    result = 0

    @remoulade.actor
    def write_loaded_at():
        nonlocal result
        result += 1

    stub_broker.declare_actor(write_loaded_at)

    scheduler.schedule = [
        ScheduledJob(
            actor_name="write_loaded_at",
            iso_weekday=datetime.datetime.now().isoweekday(),
        )
    ]
    write_loaded_at.send, event = mock_func(write_loaded_at.send)
    scheduler_thread.start()
    event.wait(2)
    stub_broker.join(write_loaded_at.queue_name)
    stub_worker.join()

    # Should have ran
    assert result == 1


def test_add_job(scheduler):
    job = ScheduledJob(actor_name="do_work")
    scheduler.add_job(job)
    hashes = list(scheduler.get_redis_schedule().keys())
    assert len(hashes) == 1
    assert hashes[0] == job.get_hash()


def test_delete_job(scheduler):
    scheduler.schedule = [ScheduledJob(actor_name="do_work")]
    scheduler.sync_config()
    scheduler.delete_job(scheduler.schedule[0].get_hash())
    assert scheduler.get_redis_schedule() == {}


def test_get_scheduled_jobs(scheduler, api_client):
    scheduler.schedule = [ScheduledJob(actor_name="do_work")]
    scheduler.sync_config()
    res = api_client.get("/scheduled/jobs")
    assert res.status_code == 200
    assert res.json == {
        "result": [
            {
                "hash": scheduler.schedule[0].get_hash(),
                "actor_name": "do_work",
                "args": [],
                "daily_time": None,
                "enabled": True,
                "interval": 86400,
                "iso_weekday": None,
                "kwargs": {},
                "last_queued": None,
                "tz": "UTC",
            }
        ]
    }


def test_api_add_job(scheduler, api_client, do_work):
    res = api_client.post(
        "/scheduled/jobs", data=json.dumps({"actor_name": "do_work"}), content_type="application/json"
    )
    assert res.status_code == 200
    assert res.json == {
        "result": [
            {
                "hash": res.json["result"][0]["hash"],
                "actor_name": "do_work",
                "args": [],
                "daily_time": None,
                "enabled": True,
                "interval": 86400,
                "iso_weekday": None,
                "kwargs": {},
                "last_queued": None,
                "tz": "UTC",
            }
        ]
    }


def test_api_delete_job(scheduler, api_client):
    scheduler.schedule = [ScheduledJob(actor_name="do_work")]
    scheduler.sync_config()
    res = api_client.delete(f"/scheduled/jobs/{scheduler.schedule[0].get_hash()}")
    assert res.status_code == 200
    assert res.json == {"result": []}


def test_update_job(scheduler, api_client, do_work):
    scheduler.schedule = [ScheduledJob(actor_name="do_work")]
    scheduler.sync_config()
    res = api_client.put(
        f"/scheduled/jobs/{scheduler.schedule[0].get_hash()}",
        data=json.dumps({"actor_name": "do_work", "enabled": False}),
        content_type="application/json",
    )
    assert res.status_code == 200
    assert res.json == {
        "result": [
            {
                "hash": res.json["result"][0]["hash"],
                "actor_name": "do_work",
                "args": [],
                "daily_time": None,
                "enabled": False,
                "interval": 86400,
                "iso_weekday": None,
                "kwargs": {},
                "last_queued": None,
                "tz": "UTC",
            }
        ]
    }


def test_api_update_jobs(scheduler, api_client, do_work):
    scheduler.schedule = [ScheduledJob(actor_name="do_work"), ScheduledJob(actor_name="do_other_work")]
    scheduler.sync_config()
    res = api_client.put(
        "/scheduled/jobs",
        data=json.dumps(
            {
                "jobs": {
                    scheduler.schedule[0].get_hash(): {"actor_name": "do_work", "enabled": False},
                    scheduler.schedule[1].get_hash(): {"actor_name": "do_work", "enabled": False, "interval": "55"},
                }
            }
        ),
        content_type="application/json",
    )
    assert res.status_code == 200
    assert len(res.json["result"]) == 2
    for i in range(2):
        assert not res.json["result"][i]["enabled"]


def test_daily_time_wrong_interval(scheduler, api_client):
    res = api_client.post(
        "/scheduled/jobs",
        data=json.dumps({"actor_name": "do_work", "daily_time": "00:00", "interval": 1000}),
        content_type="application/json",
    )
    assert res.status_code == 400


def test_invalid_timezone(scheduler, api_client):
    res = api_client.post(
        "/scheduled/jobs",
        data=json.dumps({"actor_name": "do_work", "tz": "invalid_tz"}),
        content_type="application/json",
    )
    assert res.status_code == 400


def test_invalid_actor_name(scheduler, api_client):
    res = api_client.post(
        "/scheduled/jobs", data=json.dumps({"actor_name": "invalid_actor"}), content_type="application/json"
    )
    assert res.status_code == 400


def test_tz_aware_last_queued(scheduler, api_client, do_work):
    res = api_client.post(
        "scheduled/jobs",
        data=json.dumps({"actor_name": "do_work", "last_queued": "2020-10-10 10:00:00Z"}),
        content_type="application/json",
    )

    assert res.status_code == 400



import time
from collections import Counter
from concurrent.futures import ThreadPoolExecutor

from remoulade.rate_limits import WindowRateLimiter


def test_window_rate_limiter_limits_per_window(rate_limiter_backend):
    # Given that I have a bucket rate limiter and a call database
    limiter = WindowRateLimiter(rate_limiter_backend, "window-test", limit=2, window=1)
    calls = Counter()

    # And a function that increments keys over the span of 3 seconds
    def work():
        for _ in range(15):
            for _ in range(8):
                with limiter.acquire(raise_on_failure=False) as acquired:
                    if not acquired:
                        continue

                    calls[int(time.time())] += 1

            time.sleep(0.2)

    # If I run that function multiple times concurrently
    with ThreadPoolExecutor(max_workers=8) as e:
        futures = []
        for _ in range(8):
            futures.append(e.submit(work))

        for future in futures:
            future.result()

    # I expect between 6 and 10 calls to have been made in total
    assert 6 <= sum(calls.values()) <= 10



from functools import update_wrapper

from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from flask import request
from flask_apispec import FlaskApiSpec, use_kwargs


def validate_schema(schema):
    def decorator(func):
        def wrapper(*args, **kwargs):
            if request.is_json and request.content_length and request.content_length > 0:
                body = request.get_json()
            else:
                body = {}
            res = schema().load(body)
            return func(*args, **res, **kwargs)

        return use_kwargs(schema, apply=False)(update_wrapper(wrapper, func))

    return decorator


def add_swagger(app, routes_dict):
    app.config.update(
        {
            "APISPEC_SPEC": APISpec(
                title="Remoulade API", version="v1", plugins=[MarshmallowPlugin()], openapi_version="3.0.3"
            )
        }
    )
    api_spec = FlaskApiSpec(document_options=False)

    for name, route_list in routes_dict.items():
        for route in route_list:
            api_spec.register(route, blueprint=name if name != "main" else None)
    api_spec.init_app(app)



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
from typing import Any, Dict, Generic, TypeVar, cast

import attr

from remoulade.state import State

from .broker import get_broker
from .common import generate_unique_id
from .composition import pipeline
from .encoder import Encoder, JSONEncoder
from .errors import InvalidProgress
from .result import Result

#: The global encoder instance.
global_encoder = JSONEncoder()  # type: Encoder


def get_encoder() -> Encoder:
    """Get the global encoder object.

    Returns:
      Encoder
    """
    global global_encoder
    return global_encoder


def set_encoder(encoder: Encoder) -> None:
    """Set the global encoder object.

    Parameters:
      encoder(Encoder): The encoder instance to use when serializing
        messages.
    """
    global global_encoder
    global_encoder = encoder


ResultT = TypeVar("ResultT", bound=Result[Any], covariant=True)


@attr.s(frozen=True, slots=True, kw_only=True, auto_attribs=True)
class Message(Generic[ResultT]):
    """Encapsulates metadata about messages being sent to individual actors.

    Parameters:
      queue_name(str): The name of the queue the message belogns to.
      actor_name(str): The name of the actor that will receive the message.
      args(tuple): Positional arguments that are passed to the actor.
      kwargs(dict): Keyword arguments that are passed to the actor.
      options(dict): Arbitrary options passed to the broker and middleware.
      message_id(str): A globally-unique id assigned to the actor.
      message_timestamp(int): The UNIX timestamp in milliseconds
        representing when the message was first enqueued.
    """

    queue_name: str
    actor_name: str
    args: tuple = attr.field(converter=tuple)
    kwargs: Dict
    options: Dict[str, Any]
    message_id: str = attr.field(factory=generate_unique_id)
    message_timestamp: int = attr.field(factory=lambda: int(time.time() * 1000))

    def __or__(self, other) -> pipeline:
        """Combine this message into a pipeline with "other"."""
        return pipeline((self, other))

    def asdict(self):
        """Convert this message to a dictionary."""
        return attr.asdict(self, recurse=False)

    @classmethod
    def decode(cls, data):
        """Convert a bytestring to a message."""
        return cls(**global_encoder.decode(data))

    def encode(self):
        """Convert this message to a bytestring."""
        return global_encoder.encode(self.asdict())

    def copy(self, **attributes):
        """Create a copy of this message."""
        updated_options = attributes.pop("options", {})
        options = self.options.copy()
        options.update(updated_options)
        return attr.evolve(self, **attributes, options=options)

    def build(self, options: Dict[str, Any]):
        """Build message for pipeline"""
        return self.copy(options=options)

    def cancel(self) -> None:
        """Mark a message as canceled"""
        broker = get_broker()
        backend = broker.get_cancel_backend()
        backend.cancel([self.message_id])

    def set_progress(self, progress: float) -> None:
        """Set the progress of the message.
        progress(float) number between 0 and 1 inclusive

        :raises:
            InvalidProgress: when not( 0 <= progress <= 1)
        """
        if not (0 <= progress <= 1):
            raise InvalidProgress(f"Progress {progress} is not between 0 and 1.")

        broker = get_broker()
        backend = broker.get_state_backend()
        backend.set_state(State(self.message_id, progress=progress))

    @property
    def result(self) -> ResultT:
        return cast(ResultT, Result(message_id=self.message_id))

    def __str__(self) -> str:
        return f"{self.actor_name} / {self.message_id}"



import platform

import pytest

import remoulade
import remoulade.broker
from remoulade.brokers.rabbitmq import RabbitmqBroker
from remoulade.brokers.stub import StubBroker
from remoulade.middleware import AgeLimit, Middleware, ShutdownNotifications, TimeLimit

CURRENT_OS = platform.system()
skip_on_windows = pytest.mark.skipif(CURRENT_OS == "Windows", reason="test skipped on Windows")


class EmptyMiddleware(Middleware):
    pass


def test_broker_uses_rabbitmq_if_not_set():
    # Given that no global broker is set
    remoulade.broker.global_broker = None

    # If I try to get the global broker
    # I expect a ValueError to be raised
    with pytest.raises(ValueError) as e:
        remoulade.get_broker()

    assert str(e.value) == "Broker not found, are you sure you called set_broker(broker) ?"


def test_change_broker(stub_broker):
    # Given that some actors
    @remoulade.actor
    def add(x, y):
        return x + y

    @remoulade.actor
    def modulo(x, y):
        return x % y

    # And these actors are declared
    stub_broker.declare_actor(add)
    stub_broker.declare_actor(modulo)

    # Given a new broker
    new_broker = StubBroker()

    remoulade.change_broker(new_broker)

    # I expect them to have the same actors
    assert stub_broker.actors == new_broker.actors


def test_declare_actors(stub_broker):
    # Given that some actors
    @remoulade.actor
    def add(x, y):
        return x + y

    @remoulade.actor
    def modulo(x, y):
        return x % y

    actors = [add, modulo]
    remoulade.declare_actors(actors)

    assert set(stub_broker.actors.values()) == set(actors)


def test_declare_actors_no_broker():
    remoulade.broker.global_broker = None
    with pytest.raises(ValueError):
        remoulade.declare_actors([])


def test_middleware_is_inserted_correctly():
    broker = StubBroker(middleware=[])

    assert len(broker.middleware) == 0

    broker.add_middleware(AgeLimit())

    assert len(broker.middleware) == 1

    broker.add_middleware(ShutdownNotifications())

    assert len(broker.middleware) == 2

    broker.add_middleware(TimeLimit())

    assert len(broker.middleware) == 3
    assert isinstance(broker.middleware[1], TimeLimit)


def test_custom_middleware_is_appended(stub_broker):
    class TestMiddleware(Middleware):
        pass

    middleware_count = len(stub_broker.middleware)

    stub_broker.add_middleware(TestMiddleware())

    assert len(stub_broker.middleware) == middleware_count + 1
    assert isinstance(stub_broker.middleware[middleware_count], TestMiddleware)


def test_can_instantiate_brokers_without_middleware():
    # Given that I have an empty list of middleware
    # When I pass that to the RMQ Broker
    broker = RabbitmqBroker(middleware=[])

    # Then I should get back a broker with not middleware
    assert not broker.middleware


def test_duplicate_middleware(stub_broker):
    # Given that I have a broker with the default middleware
    middleware_count = len(stub_broker.middleware)

    # When I add a middleware that is already added with options
    stub_broker.add_middleware(AgeLimit(max_age=2))

    # Then I should have the same amount of middleware, and the Age Limit middleware should be the new one
    assert len(stub_broker.middleware) == middleware_count
    assert stub_broker.get_middleware(AgeLimit).max_age == 2


def test_remove_middleware(stub_broker):
    # Given that I have a broker with multiple middlewares
    stub_broker.add_middleware(AgeLimit())
    stub_broker.add_middleware(ShutdownNotifications())
    stub_broker.add_middleware(TimeLimit())
    middleware_count = len(stub_broker.middleware)

    # When I add a middleware that is already added with options
    stub_broker.remove_middleware(ShutdownNotifications)

    # Then I should have the right amount of middleware, and the ShutdownNotifications middleware should not be there
    assert len(stub_broker.middleware) == middleware_count - 1
    assert stub_broker.get_middleware(ShutdownNotifications) is None



from flask import Blueprint
from flask_apispec import doc, marshal_with
from marshmallow import Schema, ValidationError, fields, validate, validates_schema
from werkzeug.exceptions import NotFound

from remoulade import get_broker
from remoulade.state import State, StateStatusesEnum
from remoulade.state.backends import PostgresBackend

from .apispec import validate_schema

messages_bp = Blueprint("messages", __name__, url_prefix="/messages")


class DeleteSchema(Schema):
    """
    Class to validate delete body data in /messages/states
    """

    max_age = fields.Int(allow_none=True)
    not_started = fields.Bool(load_default=False)


class StatesParamsSchema(Schema):
    """
    Class to validate the state search parameters
    """

    sort_column = fields.Str(
        allow_none=True,
        validate=validate.OneOf([field for field in State._fields if field not in ["args", "kwargs", "options"]]),
    )
    sort_direction = fields.Str(allow_none=True, validate=validate.OneOf(["asc", "desc"]))
    size = fields.Int(allow_none=True, validate=validate.Range(min=1, max=1000))
    offset = fields.Int(load_default=0)
    selected_actors = fields.List(fields.String, allow_none=True)
    selected_statuses = fields.List(
        fields.String(validate=validate.OneOf([status.name for status in StateStatusesEnum])),
        allow_none=True,
    )
    selected_message_ids = fields.List(fields.String, allow_none=True)
    selected_composition_ids = fields.List(fields.String, allow_none=True)
    start_datetime = fields.DateTime(allow_none=True)
    end_datetime = fields.DateTime(allow_none=True)

    @validates_schema
    def validate_sort(self, data, **kwargs):
        if data.get("sort_direction") and data.get("sort_column") is None:
            raise ValidationError("sort_column is not defined")


class StateSchema(Schema):
    message_id = fields.Str()
    status = fields.Str(allow_none=True)
    actor_name = fields.Str(allow_none=True)
    args = fields.List(fields.Raw(), allow_none=True)
    kwargs = fields.Dict(keys=fields.Str(), values=fields.Raw(), allow_none=True)
    options = fields.Dict(keys=fields.Str(), values=fields.Raw(), allow_none=True)
    progress = fields.Float(allow_none=True)
    priority = fields.Int(allow_none=True)
    enqueued_datetime = fields.Str(allow_none=True)
    started_datetime = fields.Str(allow_none=True)
    end_datetime = fields.Str(allow_none=True)
    queue_name = fields.Str(allow_none=True)
    composition_id = fields.Str(allow_none=True)


class StatesResponseSchema(Schema):
    data = fields.List(fields.Nested(StateSchema))
    count = fields.Int()


@messages_bp.route("/states", methods=["POST"])
@doc(tags=["state"])
@marshal_with(StatesResponseSchema)
@validate_schema(StatesParamsSchema)
def get_states(**kwargs):
    backend = get_broker().get_state_backend()
    data = [state.as_dict() for state in backend.get_states(**kwargs)]
    StatesResponseSchema().load({"data": data, "count": backend.get_states_count(**kwargs)})
    return {"data": data, "count": backend.get_states_count(**kwargs)}


@messages_bp.route("/states", methods=["DELETE"])
@doc(tags=["state"])
@validate_schema(DeleteSchema)
def clean_states(**kwargs):
    backend = get_broker().get_state_backend()
    if not isinstance(backend, PostgresBackend):
        return {"error": "deleting states is only supported by the PostgresBackend"}, 400
    get_broker().get_state_backend().clean(**kwargs)
    return {"result": "ok"}


@messages_bp.route("/states/<message_id>")
@doc(tags=["state"])
@marshal_with(StateSchema)
def get_state(message_id):
    backend = get_broker().get_state_backend()
    data = backend.get_state(message_id)
    if data is None:
        raise NotFound(f"message_id = {message_id} does not exist")
    return data.as_dict()


messages_routes = [get_states, clean_states, get_state]



from typing import Any

from .middleware import Middleware


class CatchError(Middleware):
    """Middleware that lets you enqueue another actor or message on message failure.

    Parameters:
      on_failure(Message|Actor|str): A Message, Actor or Actor name to enqueue on failure.
    """

    @property
    def actor_options(self):
        return {"on_failure"}

    def after_process_message(self, broker, message, *, result=None, exception=None):
        from .. import Message

        if message.failed:
            on_failure = self.get_option("on_failure", broker=broker, message=message)
            if isinstance(on_failure, str):
                actor = broker.get_actor(on_failure)
                actor.send(message.actor_name, type(exception).__name__, message.args, message.kwargs)
            elif isinstance(on_failure, dict):
                on_failure_message = Message[Any](**on_failure)
                on_failure_message = on_failure_message.copy(
                    args=[message.actor_name, type(exception).__name__, message.args, message.kwargs]
                )
                broker.enqueue(on_failure_message)

    def update_options_before_create_message(self, options, broker, actor_name):
        from .. import Message
        from ..actor import Actor

        on_failure = options.get("on_failure")
        if isinstance(on_failure, Message):
            options["on_failure"] = on_failure.asdict()
        elif isinstance(on_failure, Actor):
            options["on_failure"] = on_failure.actor_name
        elif on_failure is not None and not isinstance(on_failure, str):
            raise TypeError(f"on_failure must be an Message, an Actor or a string, got {type(on_failure)} instead")

        return options



import pytest

import remoulade
from remoulade import RemouladeError, Worker, group
from remoulade.cancel import Cancel
from remoulade.results import Results
from remoulade.results.backends import LocalBackend


def test_local_broker_cannot_have_non_local_backend(local_broker, result_backend):
    # Which is not a LocalBackend
    assert not isinstance(result_backend, LocalBackend)

    # Cannot be used with LocalBroker
    with pytest.raises(RuntimeError):
        local_broker.add_middleware(Results(backend=result_backend))


def test_local_broker_get_result_in_message(local_broker, local_result_backend):
    local_broker.add_middleware(Results(backend=local_result_backend))

    # Given that I have an actor that stores its results
    @remoulade.actor(store_results=True)
    def do_work():
        return 1

    # And this actor is declared
    local_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # I should get the right result
    assert message.result.get() == 1


def test_local_broker_with_pipes(local_broker, local_result_backend):
    local_broker.add_middleware(Results(backend=local_result_backend))

    # Given that I have an actor that stores its results
    @remoulade.actor(store_results=True)
    def add(a, b):
        return a + b

    # And this actor is declared
    local_broker.declare_actor(add)

    # When I run a pipe
    pipe = add.message(1, 2) | add.message(3)
    pipe.run()

    # I should get the right result
    assert pipe.result.get() == 6


def test_local_broker_with_groups(local_broker, local_result_backend):
    local_broker.add_middleware(Results(backend=local_result_backend))

    # Given that I have an actor that stores its results
    @remoulade.actor(store_results=True)
    def add(a, b):
        return a + b

    # And this actor is declared
    local_broker.declare_actor(add)

    # When I run a group
    g = group([add.message(1, 2), add.message(3, 4), add.message(4, 5)])
    g.run()

    assert list(g.results.get()) == [3, 7, 9]


def test_local_broker_forget(local_broker, local_result_backend):
    local_broker.add_middleware(Results(backend=local_result_backend))

    # Given that I have an actor that stores its results
    @remoulade.actor(store_results=True)
    def do_work():
        return 1

    # And this actor is declared
    local_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # I if forget the result
    assert message.result.get(forget=True) == 1

    # the result is forgotten
    assert message.result.get() is None


def test_local_broker_cancel(local_broker, stub_cancel_backend):
    local_broker.add_middleware(Cancel(backend=stub_cancel_backend))

    @remoulade.actor()
    def do_work():
        return 1

    # And this actor is declared
    local_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.send()

    # I can call cancel
    message.cancel()


def test_local_broker_worker_forbidden(local_broker):
    with pytest.raises(RemouladeError):
        Worker(broker=local_broker)


def test_local_raise_error(local_broker):
    # Given that I have an actor that raise an exception
    @remoulade.actor()
    def raise_error():
        raise ValueError()

    local_broker.declare_actor(raise_error)

    # When I send that actor a message, a get an exception
    with pytest.raises(ValueError):
        raise_error.send()



from subprocess import PIPE

fakebroker = object()


def test_cli_fails_to_start_given_an_invalid_broker_name(start_cli):
    # Given that this module doesn't define a broker called "idontexist"
    # When I start the cli and point it at that broker
    proc = start_cli("tests.test_cli:idontexist", stdout=PIPE, stderr=PIPE)
    proc.wait(5)

    # Then the process return code should be 2
    assert proc.returncode == 2


def test_cli_fails_to_start_given_an_invalid_broker_instance(start_cli):
    # Given that this module defines a "fakebroker" variable that's not a Broker
    # When I start the cli and point it at that broker
    proc = start_cli("tests.test_cli:fakebroker", stdout=PIPE, stderr=PIPE)
    proc.wait(5)

    # Then the process return code should be 2
    assert proc.returncode == 2



import time
from threading import Event
from unittest import mock
from unittest.mock import Mock, call

import pytest

import remoulade
from remoulade import ActorNotFound, Message, QueueJoinTimeout, Worker
from remoulade.common import current_millis
from remoulade.errors import MessageNotDelivered
from remoulade.middleware import CurrentMessage
from remoulade.results import Results


def test_rabbitmq_actors_can_be_sent_messages(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a database
    database = {}

    # And an actor that can write data to that database
    @remoulade.actor
    def put(key, value):
        database[key] = value

    # And this actor is declared
    rabbitmq_broker.declare_actor(put)

    # If I send that actor many async messages
    for i in range(100):
        assert put.send("key-%d" % i, i)

    # And I give the workers time to process the messages
    rabbitmq_broker.join(put.queue_name)
    rabbitmq_worker.join()

    # I expect the database to be populated
    assert len(database) == 100


def test_rabbitmq_queues_created_lazily(rabbitmq_broker):
    # Given that rabbitMQ has no open connection
    rabbitmq_broker.close()

    # Given that I have an actor
    @remoulade.actor
    def add(a, b):
        return a + b

    # And this actor is declared
    rabbitmq_broker.declare_actor(add)

    # queue_name should be in prepared_queues
    assert add.queue_name in rabbitmq_broker.queues

    # nothing is sent so RabbitMQ before sending a message
    assert rabbitmq_broker._connection is None

    # If I send that actor an async message
    add.send(1, 2)

    # RabbitMQ is connected to
    assert rabbitmq_broker._connection is not None


def test_rabbitmq_actors_retry_with_backoff_on_failure(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a database
    failure_time, success_time = None, None
    succeeded = Event()

    # And an actor that fails the first time it's called
    @remoulade.actor(max_retries=3, min_backoff=1000, max_backoff=5000)
    def do_work():
        nonlocal failure_time, success_time
        if not failure_time:
            failure_time = current_millis()
            raise RuntimeError("First failure.")
        else:
            success_time = current_millis()
            succeeded.set()

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_work)

    # If I send it a message
    do_work.send()

    # Then wait for the actor to succeed
    succeeded.wait(timeout=30)

    # I expect backoff time to have passed between sucesss and failure
    assert 800 <= success_time - failure_time <= 1200


def test_rabbitmq_actors_retry_with_queue_escalation_on_failure(rabbitmq_broker, rabbitmq_worker):
    for queue_name in ["default.high", "default.medium"]:
        rabbitmq_broker.declare_queue(queue_name)

    incoming_queues = []
    escalation_queue_mapping_arg = {"default": "default.medium", "default.medium": "default.high"}

    @remoulade.actor(max_retries=4, escalation_queue_mapping=escalation_queue_mapping_arg, max_backoff=1)
    def do_work():
        nonlocal incoming_queues
        msg = CurrentMessage.get_current_message()
        incoming_queues.append(msg.queue_name)
        raise RuntimeError("Failure")

    rabbitmq_broker.declare_actor(do_work)

    do_work.send()

    rabbitmq_broker.join(do_work.queue_name, min_successes=40)
    rabbitmq_worker.join()

    # Queue should be escalated only once
    assert incoming_queues == ["default", "default.medium", "default.medium", "default.medium", "default.medium"]


def test_rabbitmq_actors_retry_without_queue_escalation(rabbitmq_broker, rabbitmq_worker):
    for queue_name in ["default.high", "default.medium"]:
        rabbitmq_broker.declare_queue(queue_name)

    incoming_queues = []
    escalation_queue_mapping_arg = {}

    @remoulade.actor(max_retries=4, escalation_queue_mapping=escalation_queue_mapping_arg, max_backoff=1)
    def do_work():
        nonlocal incoming_queues
        msg = CurrentMessage.get_current_message()
        incoming_queues.append(msg.queue_name)
        raise RuntimeError("Failure")

    rabbitmq_broker.declare_actor(do_work)

    do_work.send()

    rabbitmq_broker.join(do_work.queue_name, min_successes=40)
    rabbitmq_worker.join()

    # queue should stay the same
    assert incoming_queues == ["default"] * 5


def test_rabbitmq_actors_retry_with_priority_elevation_on_failure(rabbitmq_broker, rabbitmq_worker):
    incoming_priorities = []

    @remoulade.actor(max_retries=4, increase_priority_on_retry=True, max_backoff=1)
    def do_work():
        nonlocal incoming_priorities
        msg = CurrentMessage.get_current_message()
        incoming_priorities.append(msg._rabbitmq_message.priority)
        raise RuntimeError("Failure")

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_work)
    # If I send it a message
    do_work.send()

    # Then join on the queue
    rabbitmq_broker.join(do_work.queue_name, min_successes=40)
    rabbitmq_worker.join()

    # I expect the priority to increase once
    assert incoming_priorities == [0, 1, 1, 1, 1]


def test_rabbitmq_actors_retry_without_priority_elevation(rabbitmq_broker, rabbitmq_worker):
    incoming_priorities = []

    @remoulade.actor(max_retries=4, increase_priority_on_retry=False, max_backoff=1)
    def do_work():
        nonlocal incoming_priorities
        msg = CurrentMessage.get_current_message()
        incoming_priorities.append(msg._rabbitmq_message.priority)
        raise RuntimeError("Failure")

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_work)
    # If I send it a message
    do_work.send()

    # Then join on the queue
    rabbitmq_broker.join(do_work.queue_name, min_successes=40)
    rabbitmq_worker.join()

    # I expect the priority to stay none for all attempts
    assert incoming_priorities == [0] * 5


def test_rabbitmq_actors_can_retry_multiple_times(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a database
    attempts = []

    # And an actor that fails 3 times then succeeds
    @remoulade.actor(max_retries=3, max_backoff=1)
    def do_work():
        attempts.append(1)
        if sum(attempts) < 4:
            raise RuntimeError("Failure #%d" % sum(attempts))

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_work)

    # If I send it a message
    do_work.send()

    # Then join on the queue
    rabbitmq_broker.join(do_work.queue_name, min_successes=40)
    rabbitmq_worker.join()

    # I expect it to have been attempted 4 times
    assert sum(attempts) == 4


def test_rabbitmq_actors_can_have_their_messages_delayed(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a database
    start_time, run_time = current_millis(), None

    # And an actor that records the time it ran
    @remoulade.actor
    def record():
        nonlocal run_time
        run_time = current_millis()

    # And this actor is declared
    rabbitmq_broker.declare_actor(record)

    # If I send it a delayed message
    record.send_with_options(delay=1000)

    # Then join on the queue
    rabbitmq_broker.join(record.queue_name)
    rabbitmq_worker.join()

    # I expect that message to have been processed at least delayed milliseconds later
    assert run_time - start_time >= 1000


def test_rabbitmq_actors_can_delay_messages_independent_of_each_other(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a database
    results = []

    # And an actor that appends a number to the database
    @remoulade.actor
    def append(x):
        results.append(x)

    # And this actor is declared
    rabbitmq_broker.declare_actor(append)

    # And I send it a delayed message
    append.send_with_options(args=(1,), delay=1500)

    # And then another delayed message with a smaller delay
    append.send_with_options(args=(2,), delay=1000)

    # Then join on the queue
    rabbitmq_broker.join(append.queue_name, min_successes=20)
    rabbitmq_worker.join()

    # I expect the latter message to have been run first
    assert results == [2, 1]


def test_rabbitmq_actors_can_have_retry_limits(rabbitmq_broker, rabbitmq_worker):
    # Given that I have an actor that always fails
    @remoulade.actor(max_retries=0)
    def do_work():
        raise RuntimeError("failed")

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_work)

    # If I send it a message
    do_work.send()

    # Then join on its queue
    rabbitmq_broker.join(do_work.queue_name)
    rabbitmq_worker.join()

    # I expect the message to get moved to the dead letter queue
    _, _, xq_count = rabbitmq_broker.get_queue_message_counts(do_work.queue_name)
    assert xq_count == 1


def test_rabbitmq_messages_belonging_to_missing_actors_are_rejected(rabbitmq_broker, rabbitmq_worker):
    # Given that I have a broker without actors
    # If I send it a message
    message = Message(queue_name="some-queue", actor_name="some-actor", args=(), kwargs={}, options={})
    rabbitmq_broker.declare_queue(message.queue_name)
    with pytest.raises(ActorNotFound):
        rabbitmq_broker.enqueue(message)


def test_rabbitmq_broker_reconnects_after_enqueue_failure(rabbitmq_broker):
    # Given that I have an actor
    @remoulade.actor
    def do_nothing():
        pass

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_nothing)

    # If I close my connection
    rabbitmq_broker.connection.close()

    # Then send my actor a message
    # I expect the message to be enqueued
    assert do_nothing.send()

    # And the connection be reopened
    assert rabbitmq_broker.connection.is_open


@mock.patch("amqpstorm.basic.Basic.publish")
@pytest.mark.confirm_delivery(True)
def test_rabbitmq_broker_retry_to_enqueue_message(publish, rabbitmq_broker):
    # Given that I have an actor
    @remoulade.actor
    def do_nothing():
        pass

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_nothing)

    publish.return_value = False
    with pytest.raises(MessageNotDelivered):
        assert do_nothing.send()

    assert publish.call_count == 6


def test_rabbitmq_connections_can_be_deleted_multiple_times(rabbitmq_broker):
    del rabbitmq_broker.connection
    del rabbitmq_broker.connection


def test_rabbitmq_channels_can_be_deleted_multiple_times(rabbitmq_broker):
    rabbitmq_broker.clear_channel_pools()
    rabbitmq_broker.clear_channel_pools()


def test_rabbitmq_consumers_ignore_unknown_messages_in_ack_and_nack(rabbitmq_broker):
    # Given that I have a RabbitmqConsumer
    consumer = rabbitmq_broker.consume("default")

    # If I attempt to ack a Message that wasn't consumed off of it
    # I expect nothing to happen
    assert consumer.ack(Mock(_tag=1)) is None

    # Likewise for nack
    assert consumer.nack(Mock(_tag=1)) is None


def test_rabbitmq_broker_can_join_with_timeout(rabbitmq_broker, rabbitmq_worker):
    # Given that I have an actor that takes a long time to run
    @remoulade.actor
    def do_work():
        time.sleep(1)

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_work)

    # When I send that actor a message
    do_work.send()

    # And join on its queue with a timeout
    # Then I expect a QueueJoinTimeout to be raised
    with pytest.raises(QueueJoinTimeout):
        rabbitmq_broker.join(do_work.queue_name, timeout=500)


def test_rabbitmq_broker_can_flush_queues(rabbitmq_broker):
    # Given that I have an actor
    @remoulade.actor
    def do_work():
        pass

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_work)

    # When I send that actor a message
    do_work.send()

    # And then tell the broker to flush all queues
    rabbitmq_broker.flush_all()

    # And then join on the actors's queue
    # Then it should join immediately
    assert rabbitmq_broker.join(do_work.queue_name, min_successes=1, timeout=200) is None


def test_rabbitmq_broker_can_enqueue_messages_with_priority(rabbitmq_broker):
    max_priority = 10
    message_processing_order = []
    queue_name = "prioritized"

    # Given that I have an actor that store priorities
    @remoulade.actor(queue_name=queue_name)
    def do_work(message_priority):
        message_processing_order.append(message_priority)

    remoulade.declare_actors([do_work])

    worker = Worker(rabbitmq_broker, worker_threads=1)
    worker.queue_prefetch = 1

    try:
        # When I send that actor messages with increasing priorities
        for priority in range(max_priority):
            do_work.send_with_options(args=(priority,), priority=priority)

        worker.start()
        # And then tell the broker to wait for all messages
        rabbitmq_broker.join(queue_name, timeout=5000)
        worker.join()

        # I expect the stored priorities to be saved in decreasing order
        assert message_processing_order == list(reversed(range(max_priority)))
    finally:
        worker.stop()


def test_rabbitmq_broker_disable_delivery_confirmation(rabbitmq_broker, rabbitmq_worker, stub_result_backend):
    rabbitmq_broker.add_middleware(Results(backend=stub_result_backend))

    @remoulade.actor(store_results=True)
    def do_work():
        return 1

    rabbitmq_broker.declare_actor(do_work)

    message = do_work.send_with_options(confirm_delivery=False)

    # Then join on the queue
    rabbitmq_broker.join(do_work.queue_name)
    rabbitmq_worker.join()

    assert message.result.get() == 1


def test_rabbitmq_broker_can_use_transactions(rabbitmq_broker, rabbitmq_worker):
    call_count = 0

    @remoulade.actor()
    def do_work():
        nonlocal call_count
        call_count += 1

    rabbitmq_broker.declare_actor(do_work)

    # check than a message is executed
    with rabbitmq_broker.tx():
        do_work.send()

    # then enqueue message but raise in the transaction
    with pytest.raises(ValueError):
        with rabbitmq_broker.tx():
            [do_work.send() for _ in range(10)]
            raise ValueError()

    # Then join on the queue
    rabbitmq_broker.join(do_work.queue_name)
    rabbitmq_worker.join()

    # messages have been rollback (but not the one outside the transaction)
    assert call_count == 1


@mock.patch("remoulade.brokers.rabbitmq.RabbitmqBroker._get_channel")
@pytest.mark.confirm_delivery(True)
@pytest.mark.group_transaction(True)
def test_rabbitmq_broker_delivery_confirmation_and_group_transaction(mocked_channel, rabbitmq_broker, rabbitmq_worker):
    @remoulade.actor()
    def do_work():
        return 1

    rabbitmq_broker.declare_actor(do_work)

    do_work.send()
    assert mocked_channel.call_count == 1
    assert mocked_channel.call_args == call(True)

    remoulade.group([do_work.message()]).run()
    assert mocked_channel.call_count == 2
    assert mocked_channel.call_args == call(False)



import os
from typing import Optional
from urllib.parse import urlparse

import redis
import redis.asyncio as redis_async


def redis_client(url: Optional[str], **parameters):
    """"""
    if url:
        url_parsed = urlparse(url)
        if url_parsed.scheme == "sentinel":
            sentinel = redis.Sentinel(
                [(url_parsed.hostname, url_parsed.port)],
                sentinel_kwargs={"password": url_parsed.password},
            )
            return sentinel.master_for(
                service_name=os.path.normpath(url_parsed.path).split("/")[1], password=url_parsed.password
            )
        else:
            parameters["connection_pool"] = redis.ConnectionPool.from_url(url)
            return redis.Redis(**parameters)


def async_redis_client(url: Optional[str], **parameters):
    """"""
    if url:
        url_parsed = urlparse(url)
        if url_parsed.scheme == "sentinel":
            sentinel = redis_async.Sentinel(
                [(url_parsed.hostname, url_parsed.port)],
                sentinel_kwargs={"password": url_parsed.password},
            )
            return sentinel.master_for(
                service_name=os.path.normpath(url_parsed.path).split("/")[1], password=url_parsed.password
            )
        else:
            parameters["connection_pool"] = redis_async.ConnectionPool.from_url(url)
            return redis_async.Redis(**parameters)



import pytest

import remoulade


def test_generic_actors_can_be_defined(stub_broker):
    # Given that I've subclassed GenericActor
    class Add(remoulade.GenericActor):
        def perform(self, x, y):
            return x + y

    # Then Add.__actor__ should be an instance of Actor
    assert isinstance(Add.__actor__, remoulade.Actor)

    # And it should be callable
    assert Add(1, 2) == 3


def test_generic_actors_can_be_assigned_options(stub_broker):
    # Given that I've subclassed GenericActor
    class Add(remoulade.GenericActor):
        # When I set its max_retries value to 32
        class Meta:
            max_retries = 32

        def perform(self, x, y):
            return x + y

    # Then the resulting actor should have that option set
    assert Add.options["max_retries"] == 32


def test_generic_actors_raise_not_implemented_if_perform_is_missing(stub_broker):
    # Given that I've subclassed GenericActor without implementing perform
    class Foo(remoulade.GenericActor):
        pass

    # When I call that actor
    # Then a NotImplementedError should be raised
    with pytest.raises(NotImplementedError):
        Foo()


def test_generic_actors_can_be_abstract(stub_broker, stub_worker):
    # Given that I have a calls database
    calls = set()

    # And I've subclassed GenericActor
    class BaseTask(remoulade.GenericActor):
        # When I set abstract to True
        class Meta:
            abstract = True
            queue_name = "tasks"

        def get_task_name(self):
            raise NotImplementedError

        def perform(self):
            calls.add(self.get_task_name())

    # Then BaseTask should not be an Actor
    assert not isinstance(BaseTask, remoulade.Actor)
    assert BaseTask.__actor__ is None
    with pytest.raises(AttributeError):
        BaseTask.not_existing_attr

    # When I subclass BaseTask
    class FooTask(BaseTask):
        def get_task_name(self):
            return "Foo"

    class BarTask(BaseTask):
        def get_task_name(self):
            return "Bar"

    # Then both subclasses should be actors
    # And they should inherit the parent's meta
    assert isinstance(FooTask.__actor__, remoulade.Actor)
    assert isinstance(BarTask.__actor__, remoulade.Actor)
    assert FooTask.queue_name == BarTask.queue_name == "tasks"

    # If i declare them
    stub_broker.declare_actor(BarTask.__actor__)
    stub_broker.declare_actor(FooTask.__actor__)

    # When I send both actors a message
    # And wait for them to get processed
    FooTask.send()
    BarTask.send()
    stub_broker.join(queue_name=BaseTask.Meta.queue_name)
    stub_worker.join()

    # Then my calls database should contain both task names
    assert calls == {"Foo", "Bar"}


def test_generic_actors_can_have_class_attributes(stub_broker):
    # Given a generic actor with class attributes
    class DoSomething(remoulade.GenericActor):
        STATUS_RUNNING = "running"
        STATUS_DONE = "done"

    # When I access one of it class attributes
    # Then I should get back that attribute's value
    assert DoSomething.STATUS_DONE == "done"


def test_generic_actors_can_inherite_meta(stub_broker, stub_worker):
    # Given that I have a calls database
    calls = set()

    # And I've subclassed GenericActor
    class BaseTask(remoulade.GenericActor):
        # When I set abstract to True
        class Meta:
            abstract = True
            retries = 10
            queue_name = "tasks"

        def perform(self):
            calls.add(self.get_task_name())

    # When I subclass BaseTask
    class FooTask(BaseTask):
        class Meta(BaseTask.Meta):
            retries = 20

    # Then both subclasses should be actors
    # And they should inherit the parent's meta
    assert isinstance(FooTask.__actor__, remoulade.Actor)
    assert FooTask.queue_name == "tasks"
    assert FooTask.options["retries"] == 20



import logging

from ..errors import RateLimitExceeded
from ..logging import get_logger
from .middleware import Middleware


class WorkerThreadLogging(Middleware):
    def __init__(self):
        self.logger = get_logger(__name__, type(self))

    def build_extra(self, message, max_input_size: int = 1000):
        return {
            **message.options.get("logging_metadata", {}),
            "message_id": message.message_id,
            "input": {"args": str(message.args)[:max_input_size], "kwargs": str(message.kwargs)[:max_input_size]},
        }

    def before_actor_execution(self, broker, message):
        self.logger.info("Started Actor %s", message, extra=self.build_extra(message))

    def after_actor_execution(self, broker, message, *, runtime=0):
        extra = self.build_extra(message)
        extra["runtime"] = runtime
        self.logger.info("Finished Actor %s after %.02fms.", message, runtime, extra=extra)

    def before_process_message(self, broker, message):
        self.logger.debug(
            "Received message %s with id %r.", message, message.message_id, extra=self.build_extra(message)
        )

    def after_process_message(self, broker, message, *, result=None, exception=None):
        if exception is None:
            return
        if isinstance(exception, RateLimitExceeded):
            self.logger.warning(
                "Rate limit exceeded in message %s: %s.", message, exception, extra=self.build_extra(message)
            )
        else:
            self.logger.log(
                logging.ERROR if message.failed else logging.WARNING,
                "Failed to process message %s with unhandled %s",
                message,
                exception.__class__.__name__,
                exc_info=True,
                extra=self.build_extra(message, 5000),
            )

    def after_skip_message(self, broker, message):
        self.logger.warning("Message %s was skipped.", message, extra=self.build_extra(message))

    def after_message_canceled(self, broker, message):
        self.logger.warning("Message %s has been canceled", message, extra=self.build_extra(message))



import remoulade
from remoulade.worker import build_extra

from .common import get_logs, worker


def test_workers_dont_register_queues_that_arent_whitelisted(stub_broker):
    # Given that I have a worker object with a restricted set of queues
    with worker(stub_broker, queues={"a", "b"}) as stub_worker:
        # When I try to register a consumer for a queue that hasn't been whitelisted
        stub_broker.declare_queue("c")
        stub_broker.declare_queue("c.DQ")

        # Then a consumer should not get spun up for that queue
        assert "c" not in stub_worker.consumers
        assert "c.DQ" not in stub_worker.consumers


def test_build_extra(stub_broker):
    # Given that I have an actor
    @remoulade.actor
    def do_work():
        return 1

    # and the actor is declared
    stub_broker.declare_actor(do_work)

    # I build a message
    message = do_work.message()

    # I build the extra
    extra = build_extra(message)

    # I expect the extra to look like this
    assert extra == {"message_id": message.message_id, "input": {"args": "()", "kwargs": "{}"}}


def test_build_extra_metadata(stub_broker):
    # Given that I have an actor
    @remoulade.actor
    def do_work():
        return 1

    # and the actor is declared
    stub_broker.declare_actor(do_work)

    # I build a message with logging_metadata
    message = do_work.message_with_options(logging_metadata={"id": "1"})

    # I build the extra
    extra = build_extra(message)

    # I expect the find the logging_metadata in the extra
    assert extra == {"id": "1", "message_id": message.message_id, "input": {"args": "()", "kwargs": "{}"}}


def test_build_extra_override(stub_broker):
    # Given that I have an actor
    @remoulade.actor
    def do_work():
        return 1

    # and the actor is declared
    stub_broker.declare_actor(do_work)

    # I build a message with logging_metadata that has a key already used in the extra
    message = do_work.message_with_options(logging_metadata={"message_id": "1"})

    # I build the extra
    extra = build_extra(message)

    # I expect the 'message_id' value to be the message.message_id
    assert extra == {"message_id": message.message_id, "input": {"args": "()", "kwargs": "{}"}}


def test_logging_metadata_logs(stub_broker, stub_worker, caplog):
    # Given that I have an actor
    @remoulade.actor
    def do_work():
        return 1

    # and the actor is declared
    stub_broker.declare_actor(do_work)

    # I build a message with logging_metadata
    do_work.send_with_options(logging_metadata={"id": "1"})

    # And join on the queue
    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # I expect to find the metadata in the logs extra
    records = get_logs(caplog, "Started Actor")
    assert len(records) == 1
    assert records[0].levelname == "INFO"
    assert records[0].__dict__["id"] == "1"



from unittest.mock import patch

import pytest

import remoulade
from remoulade import group, pipeline
from remoulade.cancel import Cancel
from remoulade.errors import NoCancelBackend


@pytest.mark.parametrize("cancel", [True, False])
def test_actors_can_be_canceled(stub_broker, stub_worker, cancel_backend, cancel):
    # Given a cancel backend
    # And a broker with the cancel middleware
    stub_broker.add_middleware(Cancel(backend=cancel_backend))

    has_been_called = []

    # And an actor
    @remoulade.actor()
    def do_work():
        has_been_called.append(1)

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # When I send that actor a message
    message = do_work.message()

    # If I cancel the message
    if cancel:
        message.cancel()

    stub_broker.enqueue(message)

    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # It should not be called
    assert bool(has_been_called) != cancel


def test_cancellations_not_stored_forever(cancel_backend, frozen_datetime):
    # Given a cancel backend which store the cancellations 1 second
    cancel_backend.cancellation_ttl = 1

    # If we cancel some messages
    cancel_backend.cancel("a")
    # Then wait some time
    frozen_datetime.tick(delta=2)
    # And call again cancel
    cancel_backend.cancel("c")

    # the first cancellation should have been forgotten
    assert not cancel_backend.is_canceled("a", None)


@pytest.mark.parametrize("cancel", [True, False])
def test_compositions_are_canceled_on_actor_failure(stub_broker, stub_worker, cancel_backend, cancel):
    # Given a cancel backend
    # And a broker with the cancel middleware
    stub_broker.add_middleware(Cancel(backend=cancel_backend))

    # And an actor who doesn't fail
    @remoulade.actor
    def do_work(arg=None):
        return 1

    # And an actor who fails
    @remoulade.actor
    def do_fail(arg):
        raise Exception

    # And those actors are declared
    remoulade.declare_actors([do_work, do_fail])

    g = group([do_work.message(), do_fail.message() | do_work.message()], cancel_on_error=cancel)
    p = pipeline([g, do_work.message()], cancel_on_error=cancel)

    # When I group a few jobs together and run it
    with patch("remoulade.composition.generate_unique_id") as mocked_obj:
        mocked_obj.return_value = "mocked_composition_id"
        p.run()

    stub_broker.join(do_fail.queue_name)
    stub_worker.join()

    # All actors should have been canceled
    assert cancel_backend.is_canceled("", "mocked_composition_id") == cancel


def test_compositions_are_canceled_on_message_cancel(stub_broker, cancel_backend, state_middleware, api_client):
    # Given a cancel backend
    # And a broker with the cancel middleware
    stub_broker.add_middleware(Cancel(backend=cancel_backend))

    # And an actor
    @remoulade.actor
    def do_work(arg=None):
        return 1

    # And those actors are declared
    stub_broker.declare_actor(do_work)

    message_to_cancel = do_work.message()

    # And a group that I enqueue
    g = group([message_to_cancel | do_work.message(), do_work.message()], cancel_on_error=True)
    g.run()

    # When I cancel a message of this group
    api_client.post("messages/cancel/" + message_to_cancel.message_id)

    # The whole composition should be canceled
    assert cancel_backend.is_canceled("", g.group_id)


def test_cannot_cancel_on_error_if_no_cancel(stub_broker):
    # Given an actor
    @remoulade.actor()
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    with pytest.raises(NoCancelBackend):
        group((do_work.message() for _ in range(4)), cancel_on_error=True)


@pytest.mark.parametrize("with_pipeline", [True, False])
def test_cancel_pipeline_or_groups(stub_broker, stub_worker, cancel_backend, with_pipeline):
    # Given a cancel backend
    # And a broker with the cancel middleware
    stub_broker.add_middleware(Cancel(backend=cancel_backend))

    has_been_called = []

    # And an actor
    @remoulade.actor()
    def do_work():
        has_been_called.append(1)
        raise ValueError()

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    if with_pipeline:
        g = pipeline(do_work.message() for _ in range(4))
    else:
        g = group(do_work.message() for _ in range(4))

    g.cancel()
    g.run()

    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # All actors should have been canceled
    assert all(cancel_backend.is_canceled(child.message_id, None) for child in g.children)
    assert len(has_been_called) == 0


def test_composition_can_be_canceled(stub_broker, stub_worker, cancel_backend):
    # Given a cancel backend
    # And a broker with the cancel middleware
    stub_broker.add_middleware(Cancel(backend=cancel_backend))

    calls_count = 0

    # And an actor
    @remoulade.actor()
    def do_work():
        nonlocal calls_count
        calls_count += 1
        raise ValueError()

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # And a composition
    g = group([do_work.message() | do_work.message() for _ in range(2)])

    # If the composition is canceled
    cancel_backend.cancel([g.group_id])

    g.run()

    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # It messages should not have runMessageSchema
    assert calls_count == 0


def test_raise_error_if_unknown_id(stub_broker, cancel_backend, api_client):
    # Given a cancel middleware
    stub_broker.add_middleware(Cancel(backend=cancel_backend))

    # If I try to cancel a id that is not a id of a message or composition
    res = api_client.post("messages/cancel/invalid_id")

    assert res.status_code == 400



import threading
import time
from threading import Condition
from unittest import mock
from unittest.mock import call, patch

import pytest

import remoulade
from remoulade import CollectionResults, group, pipeline
from remoulade.results import ErrorStored, MessageIdsMissing, ResultMissing, Results, ResultTimeout
from remoulade.results.backends import StubBackend
from tests.common import get_logs
from tests.conftest import fast_backoff, mock_func


def test_messages_can_be_piped(stub_broker):
    # Given an actor that adds two numbers together
    @remoulade.actor
    def add(x, y):
        return x + y

    # And this actor is declared
    stub_broker.declare_actor(add)

    # When I pipe some messages intended for that actor together
    pipe = add.message(1, 2) | add.message(3) | add.message(4)

    # Then I should get back a pipeline object
    assert isinstance(pipe, pipeline)

    def filter_options(message):
        return {key: value for (key, value) in message.items() if key != "options"}

    # If I build a pipeline
    targets = pipe.build()
    assert len(targets) == 1
    first_target = targets[0]
    assert len(first_target.options["pipe_target"]) == 1
    # And each message in the pipeline should reference the next message in line
    assert filter_options(first_target.options["pipe_target"][0]) == filter_options(pipe.children[1].asdict())

    assert len(first_target.options["pipe_target"]) == 1
    second_target = first_target.options["pipe_target"][0]
    assert filter_options(second_target["options"]["pipe_target"][0]) == filter_options(pipe.children[2].asdict())

    assert len(second_target["options"]["pipe_target"]) == 1
    third_target = second_target["options"]["pipe_target"][0]
    assert "pipe_target" not in third_target["options"]


def test_pipelines_flatten_child_pipelines(stub_broker):
    # Given an actor that adds two numbers together
    @remoulade.actor
    def add(x, y):
        return x + y

    # And this actor is declared
    stub_broker.declare_actor(add)

    # When I pipe a message intended for that actor and another pipeline together
    pipe = pipeline([add.message(1, 2), add.message(3) | add.message(4), add.message(5)])

    # Then the inner pipeline should be flattened into the outer pipeline
    assert len(pipe) == 4
    assert pipe.children[0].args == (1, 2)
    assert pipe.children[1].args == (3,)
    assert pipe.children[2].args == (4,)
    assert pipe.children[3].args == (5,)


def test_pipe_ignore_message_options(stub_broker, stub_worker, result_backend):
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that return something
    @remoulade.actor(store_results=True)
    def do_nothing():
        return 0

    # Nothing should be sent to pipe ignored
    @remoulade.actor(store_results=True)
    def pipe_ignored(*args):
        assert len(args) == 0
        return 1

    # And these actors are declared
    stub_broker.declare_actor(do_nothing)
    stub_broker.declare_actor(pipe_ignored)

    pipe = do_nothing.message() | pipe_ignored.message_with_options(pipe_ignore=True)
    pipe.run()

    assert pipe.result.get(block=True) == 1


def test_pipe_ignore_actor_options(stub_broker, stub_worker, result_backend):
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that return something
    @remoulade.actor(store_results=True)
    def do_nothing():
        return 0

    # Nothing should be sent to pipe ignored
    @remoulade.actor(store_results=True, pipe_ignore=True)
    def pipe_ignored(*args):
        assert len(args) == 0
        return 1

    # And these actors are declared
    stub_broker.declare_actor(do_nothing)
    stub_broker.declare_actor(pipe_ignored)

    pipe = do_nothing.message() | pipe_ignored.message()
    pipe.run()

    assert pipe.result.get(block=True) == 1


def test_pipeline_results_can_be_retrieved(stub_broker, stub_worker, result_backend):
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that adds two numbers together and stores the result
    @remoulade.actor(store_results=True)
    def add(x, y):
        return x + y

    # And this actor is declared
    stub_broker.declare_actor(add)

    # When I pipe some messages intended for that actor together and run the pipeline
    pipe = add.message(1, 2) | (add.message(3) | add.message(4))
    pipe.run()

    # Then the pipeline result should be the sum of 1, 2, 3 and 4
    assert pipe.result.get(block=True) == 10

    # And I should be able to retrieve individual results
    assert list(pipe.results.get()) == [3, 6, 10]


def test_pipeline_results_respect_timeouts(stub_broker, stub_worker, result_backend):
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that waits some amount of time then doubles that amount
    @remoulade.actor(store_results=True)
    def wait(n):
        time.sleep(n)
        return n * 2

    # And this actor is declared
    stub_broker.declare_actor(wait)

    # When I pipe some messages intended for that actor together and run the pipeline
    pipe = wait.message(1) | wait.message() | wait.message()
    pipe.run()

    # And get the results with a lower timeout than the tasks can complete in
    # Then a ResultTimeout error should be raised
    with pytest.raises(ResultTimeout):
        for _ in pipe.results.get(block=True, timeout=1000):
            pass


def test_pipelines_expose_completion_stats(stub_broker, stub_worker, result_backend):
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that waits some amount of time
    result_backend.store_results, event_result = mock_func(result_backend.store_results)
    event_count = [threading.Event() for _ in range(4)]

    @remoulade.actor(store_results=True)
    def wait(n):
        event_count[n].wait(3)
        return n + 1

    # And this actor is declared
    stub_broker.declare_actor(wait)

    # When I pipe some messages intended for that actor together and run the pipeline
    pipe = wait.message(0) | wait.message() | wait.message() | wait.message()
    pipe.run()

    # Then every time a job in the pipeline completes, the completed_count should increase
    for count in range(0, len(pipe)):
        event_count[count].set()
        event_result.wait(2)
        event_result.clear()
        assert pipe.results.completed_count == count + 1

    # Finally, completed should be true
    assert pipe.results.completed


def test_pipelines_can_be_incomplete(stub_broker, result_backend):
    # Given that I am not running a worker
    # And I have a result backend
    stub_broker.add_middleware(Results(backend=result_backend))

    # And I have an actor that does nothing
    @remoulade.actor(store_results=True)
    def do_nothing():
        return None

    # And this actor is declared
    stub_broker.declare_actor(do_nothing)

    # And I've run a pipeline
    pipe = do_nothing.message() | do_nothing.message_with_options(pipe_ignore=True)
    pipe.run()

    # When I check if the pipeline has completed
    # Then it should return False
    assert not pipe.results.completed


@pytest.mark.parametrize("store_results", [True, False])
def test_pipelines_store_results_error(stub_broker, result_backend, stub_worker, store_results):
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # Given an actor that fail
    @remoulade.actor(store_results=store_results)
    def do_work_fail():
        raise ValueError()

    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    # And these actors are declared
    stub_broker.declare_actor(do_work_fail)
    stub_broker.declare_actor(do_work)

    # And I've run a pipeline
    g = group([do_work.message(), do_work.message(), do_work.message()])
    pipe = do_work_fail.message() | do_work.message() | g | do_work.message()
    pipe.run()

    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # I get an error
    if store_results:
        with pytest.raises(ErrorStored) as e:
            pipe.children[0].result.get()
        assert str(e.value) == "ValueError()"

    for i in [1, 3]:
        with pytest.raises(ErrorStored) as e:
            pipe.children[i].result.get()
        assert str(e.value).startswith("ParentFailed")

    for child in g.children:
        with pytest.raises(ErrorStored) as e:
            child.result.get()
        assert str(e.value).startswith("ParentFailed")


def test_groups_execute_jobs_in_parallel(stub_broker, stub_worker, result_backend):
    # Given that I have a result backend
    stub_broker.add_middleware(Results(backend=result_backend))

    # And I have an actor that sleeps for 100ms
    @remoulade.actor(store_results=True)
    def wait():
        time.sleep(0.1)

    # And this actor is declared
    stub_broker.declare_actor(wait)

    # When I group multiple of these actors together and run them
    t = time.monotonic()
    g = group([wait.message() for _ in range(5)])
    g.run()

    # group message_ids are no stored if not needed
    with pytest.raises(MessageIdsMissing):
        result_backend.get_group_message_ids(g.group_id)

    # And wait on the group to complete
    results = list(g.results.get(block=True))

    # Then the total elapsed time should be less than 500ms
    assert time.monotonic() - t <= 0.5

    # And I should get back as many results as there were jobs in the group
    assert len(results) == len(g)

    # And the group should be completed
    assert g.results.completed
    assert isinstance(g.results, CollectionResults)


def test_inner_groups_forbidden(stub_broker, stub_worker, result_backend):
    # Given that I have a result backend
    stub_broker.add_middleware(Results(backend=result_backend))

    # And I have an actor
    @remoulade.actor()
    def do_work():
        return 1

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # groups of groups are forbidden
    with pytest.raises(ValueError):
        group(group(do_work.message() for _ in range(2)) for _ in range(3))


def test_groups_can_time_out(stub_broker, stub_worker, result_backend):
    # Given that I have a result backend
    stub_broker.add_middleware(Results(backend=result_backend))

    # And I have an actor that sleeps for 300ms
    @remoulade.actor(store_results=True)
    def wait():
        time.sleep(0.3)

    # And this actor is declared
    stub_broker.declare_actor(wait)

    # When I group a few jobs together and run it
    g = group(wait.message() for _ in range(2))
    g.run()

    # And wait for the group to complete with a timeout
    # Then a ResultTimeout error should be raised
    with pytest.raises(ResultTimeout):
        g.results.wait(timeout=100)

    # And the group should not be completed
    assert not g.results.completed


def test_groups_expose_completion_stats(stub_broker, stub_worker, result_backend):
    # Given that I have a result backend
    stub_broker.add_middleware(Results(backend=result_backend))

    # And an actor that waits some amount of time
    condition = Condition()

    @remoulade.actor(store_results=True)
    def wait(n):
        time.sleep(n)
        with condition:
            condition.notify_all()
            return n

    # And this actor is declared
    stub_broker.declare_actor(wait)

    # When I group messages of varying durations together and run the group
    g = group(wait.message(n) for n in range(1, 4))
    g.run()

    # Then every time a job in the group completes, the completed_count should increase
    for count in range(1, len(g) + 1):
        with condition:
            condition.wait(5)
            time.sleep(0.1)  # give the worker time to set the result
            assert g.results.completed_count == count

    # Finally, completed should be true
    assert g.results.completed


@pytest.mark.parametrize("block", [True, False])
def test_group_forget(stub_broker, result_middleware, stub_worker, block):
    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # And I've run a group
    messages = [do_work.message() for _ in range(5)]
    g = group(messages)
    g.run()

    # If i wait for the group to be completed
    if not block:
        stub_broker.join(do_work.queue_name)
        stub_worker.join()

    # If i forget the results
    results = g.results.get(block=block, forget=True)
    assert list(results) == [42] * 5

    # All messages have been forgotten
    results = g.results.get()
    assert list(results) == [None] * 5


def test_group_wait_forget(stub_broker, result_backend, stub_worker):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    # And this actor is declared
    stub_broker.declare_actor(do_work)

    # And I've run a group
    messages = [do_work.message() for _ in range(5)]
    g = group(messages)
    g.run()

    # If i forget the results
    g.results.wait(forget=True)

    # All messages have been forgotten
    assert list(g.results.get()) == [None] * 5


def test_pipelines_with_groups(stub_broker, stub_worker, result_middleware):
    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work(a):
        return a

    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_sum(results):
        return sum(results)

    # And this actor is declared
    stub_broker.declare_actor(do_work)
    stub_broker.declare_actor(do_sum)

    # When I pipe some messages intended for that actor together and run the pipeline
    g = group([do_work.message(12), do_work.message(15)])
    pipe = g | do_sum.message()

    pipe.build()
    assert result_middleware.backend.get_group_message_ids(g.group_id) == list(g.message_ids)

    pipe.run()

    result = pipe.result.get(block=True)

    assert 12 + 15 == result

    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    # the group result has been forgotten
    assert list(g.results.get()) == [None, None]

    # the message_ids has been forgotten
    with pytest.raises(MessageIdsMissing):
        result_middleware.backend.get_group_message_ids(g.group_id)

    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def add(a, b):
        return a + b

    stub_broker.declare_actor(add)

    pipe = do_work.message(13) | group([add.message(12), add.message(15)])
    pipe.run()

    result = pipe.result.get(block=True)

    assert [13 + 12, 13 + 15] == list(result)


def test_group_one_not_finished(stub_broker, stub_worker, result_middleware):
    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        return 42

    stub_broker.declare_actor(do_work)

    messages = [do_work.send(), do_work.message(), do_work.send()]
    results = CollectionResults([m.result for m in messages])

    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    with pytest.raises(ResultMissing):
        list(results.get())
    assert messages[0].result.get() == messages[2].result.get() == 42


def test_complex_pipelines(stub_broker, stub_worker, result_middleware):
    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        return 1

    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def add(a):
        return 1 + a

    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_sum(results):
        return sum(results)

    # And this actor is declared
    stub_broker.declare_actor(do_work)
    stub_broker.declare_actor(do_sum)
    stub_broker.declare_actor(add)

    pipe = do_work.message_with_options(pipe_ignore=True) | add.message() | add.message()  # return 3 [1, 2, 3] ?
    g = group([pipe, add.message(), add.message(), do_work.message_with_options(pipe_ignore=True)])  # return [3,2,2,1]
    final_pipe = do_work.message() | g | do_sum.message() | add.message()  # return 9
    final_pipe.run()

    result = final_pipe.result.get(block=True)

    assert 9 == result


def test_pipeline_with_groups_and_pipe_ignore(stub_broker, stub_worker, result_backend):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # Given an actor that do not stores results
    @remoulade.actor()
    def do_work():
        return 1

    @remoulade.actor(store_results=True)
    def do_other_work():
        return 2

    # And this actor is declared
    stub_broker.declare_actor(do_work)
    stub_broker.declare_actor(do_other_work)

    # When I pipe a group with another actor
    pipe = group([do_work.message(), do_work.message()]) | do_other_work.message_with_options(pipe_ignore=True)
    pipe.run()

    # I don't get any error as long the second actor has pipe_ignore=True
    result = pipe.result.get(block=True)

    assert 2 == result

    # But if it don't, the pipeline cannot finish
    pipe = group([do_work.message(), do_work.message()]) | do_other_work.message()
    pipe.run()

    stub_broker.join(do_work.queue_name)
    stub_worker.join()

    with pytest.raises(ResultMissing):
        pipe.result.get()


def test_multiple_groups_pipelines(stub_broker, stub_worker, result_backend):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=result_backend))

    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_work():
        return 1

    # Given an actor that stores results
    @remoulade.actor(store_results=True)
    def do_sum(results):
        return sum(results)

    # And this actor is declared
    stub_broker.declare_actor(do_work)
    stub_broker.declare_actor(do_sum)

    pipe = pipeline(
        [group([do_work.message(), do_work.message()]), group([do_sum.message(), do_sum.message()]), do_sum.message()]
    ).run()

    result = pipe.result.get(block=True)

    assert 4 == result


def test_pipeline_does_not_continue_to_next_actor_when_message_is_marked_as_failed(stub_broker, stub_worker):
    # Given that I have an actor that fails messages
    class FailMessageMiddleware(remoulade.middleware.Middleware):
        def after_process_message(self, broker, message, *, result=None, exception=None):
            message.fail()

    stub_broker.add_middleware(FailMessageMiddleware())

    has_run = False

    @remoulade.actor
    def do_nothing():
        pass

    @remoulade.actor
    def should_never_run():
        nonlocal has_run
        has_run = True

    # And this actor is declared
    stub_broker.declare_actor(do_nothing)
    stub_broker.declare_actor(should_never_run)

    # When I pipe some messages intended for that actor together and run the pipeline
    pipe = do_nothing.message_with_options(pipe_ignore=True) | should_never_run.message()
    pipe.run()

    stub_broker.join(should_never_run.queue_name, timeout=10 * 1000)
    stub_worker.join()

    # Then the second message in the pipe should never have run
    assert not has_run


@mock.patch("remoulade.results.backend.compute_backoff", fast_backoff)
def test_retry_if_increment_group_completion_fail(stub_broker, stub_worker, caplog):
    with patch.object(StubBackend, "increment_group_completion") as mock_increment_group_completion:
        mock_increment_group_completion.side_effect = Exception("Cannot increment")
        middleware = Results(backend=StubBackend())
        stub_broker.add_middleware(middleware)

        attempts = []

        # And an actor that stores results
        @remoulade.actor(store_results=True)
        def do_work(*args):
            attempts.append(1)

        # And this actor is declared
        stub_broker.declare_actor(do_work)

        # When I send that actor a message
        (group([do_work.message(), do_work.message()]) | do_work.message()).run()

        # And wait for a result
        stub_broker.join(do_work.queue_name)
        stub_worker.join()

        # The actor has been tried 8 times (4 time each do_work and never the last one)
        assert len(attempts) == 8

        records = get_logs(caplog, "Unexpected failure in after_process_message")
        assert len(records) == 0


def test_composition_id_override(stub_broker, do_work):
    group_messages = group([pipeline([do_work.message(), do_work.message()])]).build(options={"composition_id": "id"})
    assert group_messages[0].options["composition_id"] == "id"
    assert group_messages[0].options["pipe_target"][0]["options"]["composition_id"] == "id"
    pipeline_messages = pipeline([group([do_work.message(), do_work.message()])]).build(composition_id="id")
    assert all(message.options["composition_id"] == "id" for message in pipeline_messages)


@mock.patch("remoulade.brokers.rabbitmq.RabbitmqBroker._get_channel")
@pytest.mark.confirm_delivery(True)
@pytest.mark.group_transaction(True)
def test_pipeline_with_delivery_confirmation_and_group_transaction(mocked_channel, rabbitmq_broker, rabbitmq_worker):
    @remoulade.actor(pipe_ignore=True)
    def do_work():
        return 42

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_work)

    # When I pipe some messages intended for that actor together and run the pipeline
    pipe = do_work.message() | do_work.message()
    pipe.build()

    pipe.run()
    assert mocked_channel.call_count == 1
    assert mocked_channel.call_args == call(False)

    pipe.run(transaction=False)
    assert mocked_channel.call_count == 2
    assert mocked_channel.call_args == call(True)


@mock.patch("remoulade.brokers.rabbitmq.RabbitmqBroker._get_channel")
@pytest.mark.confirm_delivery(True)
@pytest.mark.group_transaction(True)
def test_group_with_delivery_confirmation_and_group_transaction(mocked_channel, rabbitmq_broker, rabbitmq_worker):
    @remoulade.actor()
    def do_work():
        return 42

    # And this actor is declared
    rabbitmq_broker.declare_actor(do_work)

    # And I've run a group
    messages = [do_work.message()]
    g = group(messages)

    g.run()
    assert mocked_channel.call_count == 1
    assert mocked_channel.call_args == call(False)

    g.run(transaction=False)
    assert mocked_channel.call_count == 2
    assert mocked_channel.call_args == call(True)


def test_pipeline_pipe_on_error(stub_broker, stub_worker, stub_result_backend):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=stub_result_backend))

    @remoulade.actor(pipe_on_error=True, store_results=True)
    def do_raise():
        raise MemoryError("BOOM")

    @remoulade.actor(store_results=True)
    def rcv_error(data):
        return data

    # And this actor is declared
    stub_broker.declare_actor(do_raise)
    stub_broker.declare_actor(rcv_error)

    pipe = pipeline([do_raise.message(), rcv_error.message()])
    pipe.run()

    stub_broker.join(do_raise.queue_name)
    stub_broker.join(rcv_error.queue_name)
    stub_worker.join()

    assert pipe.result.get(block=True, raise_on_error=False) == "MemoryError('BOOM')"


def test_complex_pipeline_pipe_on_error(stub_broker, stub_worker, stub_result_backend):
    # Given a result backend
    # And a broker with the results middleware
    stub_broker.add_middleware(Results(backend=stub_result_backend))

    @remoulade.actor(pipe_on_error=True, store_results=True)
    def do_one():
        return 1

    @remoulade.actor(pipe_on_error=True, store_results=True)
    def do_raise():
        raise MemoryError("BOOM")

    @remoulade.actor(store_results=True)
    def rcv_error(data):
        return data

    # And this actor is declared
    stub_broker.declare_actor(do_one)
    stub_broker.declare_actor(do_raise)
    stub_broker.declare_actor(rcv_error)

    pipe = pipeline([group([do_one.message(), do_raise.message(), do_one.message()]), rcv_error.message()])
    pipe.run()

    stub_worker.join()

    assert pipe.result.get(block=True, raise_on_error=False) == [1, "MemoryError('BOOM')", 1]
